from __future__ import absolute_import, print_function

import logging
from functools import partial
import webbrowser
import calendar
import os

from kivy.uix.screenmanager import Screen, ScreenManager, SlideTransition, ScreenManagerException
from kivy.uix.label import Label
from kivy.uix.gridlayout import GridLayout
from kivy.uix.boxlayout import BoxLayout
from kivy.properties import ObjectProperty, StringProperty, BooleanProperty, ListProperty, DictProperty
from kivy.uix.recycleview.views import RecycleDataViewBehavior
from kivy.uix.button import Button
from kivy.uix.togglebutton import ToggleButton
from kivy.animation import Animation
from kivy.clock import Clock, mainthread
from kivy.uix.behaviors import FocusBehavior
from kivy.uix.recycleview.layout import LayoutSelectionBehavior
from kivy.uix.recycleboxlayout import RecycleBoxLayout
from kivy.app import App
from kivy.uix.popup import Popup
from kivy.uix.behaviors import ButtonBehavior
from kivy.uix.image import Image
from kivy.uix.modalview import ModalView
from kivy.uix.textinput import TextInput

from es_gui.apps.valuation.reporting import ValuationReport
from es_gui.resources.widgets.common import MyPopup, WarningPopup, WizardCompletePopup, TileButton, RecycleViewRow, BASE_TRANSITION_DUR, BUTTON_FLASH_DUR, ANIM_STAGGER,FADEIN_DUR, SLIDER_DUR, fade_in_animation


class ValuationWizard(Screen):
    """The main screen for the valuation wizard. This hosts the nested screen manager for the actual wizard."""
    def on_enter(self):
        ab = self.manager.nav_bar
        ab.reset_nav_bar()
        ab.set_title('Wizard')

    def on_leave(self):
        # Reset wizard to initial state by removing all screens except the first.
        self.sm.current = 'start'

        if len(self.sm.screens) > 1:
            self.sm.clear_widgets(screens=self.sm.screens[1:])


class ValuationWizardScreenManager(ScreenManager):
    """The screen manager for the valuation wizard screens."""
    def __init__(self, **kwargs):
        super(ValuationWizardScreenManager, self).__init__(**kwargs)

        self.transition = SlideTransition()
        self.add_widget(ValuationWizardStart(name='start'))


class ValuationWizardStart(Screen):
    """The starting/welcome screen for the valuation wizard."""
    def _next_screen(self):
        if not self.manager.has_screen('iso_select'):
            screen = ValuationWizardISOSelect(name='iso_select')
            self.manager.add_widget(screen)

        self.manager.transition.duration = BASE_TRANSITION_DUR
        self.manager.transition.direction = 'left'
        self.manager.current = 'iso_select'


class ValuationWizardTemplate(Screen):
    """A screen template for the valuation wizard."""
    pass


class ToggleTileButton(ToggleButton, TileButton):
    selection_animation = Animation(transition='linear', duration=BUTTON_FLASH_DUR, opacity=0) \
                          + Animation(transition='linear', duration=BUTTON_FLASH_DUR, opacity=1) \
                          + Animation(transition='linear', duration=BUTTON_FLASH_DUR, opacity=0) \
                          + Animation(transition='linear', duration=BUTTON_FLASH_DUR, opacity=1)


class WizardISOTileButton(ToggleTileButton):
    """Tile button with toggle behavior for each ISO option."""
    iso_screen = None

    def __init__(self, name, **kwargs):
        super(WizardISOTileButton, self).__init__(**kwargs)

        self._name = name

    @property
    def name(self):
        """The name of the ISO/market area."""
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    def on_press(self):
        if self.state == 'down':
            self.selection_animation.start(self)
            self.iso_screen.iso = self.name


class ValuationWizardISOSelect(Screen):
    """The ISO selection screen for the valuation wizard."""
    iso = StringProperty('')
    has_selection = BooleanProperty(False)

    def __init__(self, **kwargs):
        super(ValuationWizardISOSelect, self).__init__(**kwargs)

        WizardISOTileButton.iso_screen = self

    def on_pre_enter(self):
        while len(self.iso_select.children) > 0:
            for widget in self.iso_select.children:
                if isinstance(widget, WizardISOTileButton):
                    self.iso_select.remove_widget(widget)
        
        ISO_OPTIONS = App.get_running_app().data_manager.get_markets()

        for iso in ISO_OPTIONS:
            iso_button = WizardISOTileButton(name=iso, text=iso, group='iso')
            self.iso_select.add_widget(iso_button)

    def on_enter(self):
        Clock.schedule_once(partial(fade_in_animation, self.content), 0)

        def _fade_in_buttons(button, *args):
            anim = Animation(transition='out_expo', duration=FADEIN_DUR, opacity=1)
            anim.start(button)

        for ix, iso_button in enumerate(reversed(self.iso_select.children), start=1):
            Clock.schedule_once(partial(_fade_in_buttons, iso_button), ix*ANIM_STAGGER)

    def on_leave(self):
        Animation.stop_all(self.content, 'opacity')
        self.content.opacity = 0

        for iso_button in self.iso_select.children:
            Animation.cancel_all(iso_button)
            iso_button.opacity = 0

    def on_iso(self, instance, value):
        logging.info('ValuationWizard: ISO changed to {0}.'.format(value))

        self.has_selection = True if \
            any([button.state == 'down' for button in self.iso_select.children]) \
            else False

    def on_has_selection(self, instance, value):
        # Enables/disables the next button if an ISO is selected.
        self.next_button.disabled = not value

    def _next_screen(self, *args):
        """Adds the revenue stream selection screen if it does not exist and changes screens to it."""
        if not self.manager.has_screen('node_select'):
            screen = ValuationWizardNodeSelect(name='node_select')
            self.manager.add_widget(screen)

        self.manager.transition.duration = BASE_TRANSITION_DUR
        self.manager.transition.direction = 'left'
        self.manager.current = 'node_select'


class ValuationWizardNodeSelect(Screen):
    iso = StringProperty('')
    has_selection = BooleanProperty(False)
    node = DictProperty('')

    def __init__(self, **kwargs):
        super(ValuationWizardNodeSelect, self).__init__(**kwargs)

        NodeRecycleViewRow.node_screen = self

    def on_pre_enter(self):
        self.iso = self.manager.get_screen('iso_select').iso

    def on_enter(self):
        Clock.schedule_once(partial(fade_in_animation, self.content), 0)

    def on_leave(self):
        Animation.stop_all(self.content, 'opacity')
        self.content.opacity = 0

    def on_iso(self, instance, value):
        self.has_selection = False
        self.node_rv.deselect_all_nodes()

        try:
            data_manager = App.get_running_app().data_manager
            node_options = [{'name': node[1],
                            'nodeid': node[0]} for node in data_manager.get_nodes(value).items()]
            self.node_rv.data = node_options
            self.node_rv.unfiltered_data = node_options
        except AttributeError:
            pass

    def on_node(self, instance, value):
        logging.info('ValuationWizard: Pricing node changed to {0} (ID: {1}).'.format(self.node['name'], self.node['nodeid']))

        self.has_selection = True

    def on_has_selection(self, instance, value):
        # Enables the next button if a node is selected.
        self.next_button.disabled = not value

    def _next_screen(self):
        """Adds the selection summary screen if it does not exist and changes screens to it."""
        if not self.manager.has_screen('rev_stream_select'):
            screen = ValuationWizardRevenueSelect(name='rev_stream_select')
            self.manager.add_widget(screen)

        self.manager.transition.duration = BASE_TRANSITION_DUR
        self.manager.transition.direction = 'left'
        self.manager.current = 'rev_stream_select'


class NodeRecycleViewRow(RecycleViewRow):
    """The representation widget for node in the node selector RecycleView."""
    node_screen = None

    def on_touch_down(self, touch):
        """Add selection on touch down."""
        if super(NodeRecycleViewRow, self).on_touch_down(touch):
            return True
        if self.collide_point(*touch.pos) and self.selectable:
            return self.parent.select_with_touch(self.index, touch)

    def apply_selection(self, rv, index, is_selected):
        """Respond to the selection of items in the view."""
        self.selected = is_selected

        if is_selected:
            self.node_screen.node = rv.data[self.index]


class WizardRevToggleButton(ToggleTileButton):
    """Tile button with toggle and hover behavior for each revenue stream option."""
    rev_screen = None
    desc_box = None

    def __init__(self, desc, name, **kwargs):
        super(WizardRevToggleButton, self).__init__(**kwargs)

        self._name = name
        self._desc = desc
        self._final_pos = None

    @property
    def name(self):
        """The name of the market formulation as known to the ValuationOptimizer."""
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def desc(self):
        """A description of the market formulation."""
        return self._desc

    @desc.setter
    def desc(self, value):
        self._desc = value

    @property
    def final_pos(self):
        """The final position of the button."""
        return self._final_pos

    @final_pos.setter
    def final_pos(self, value):
        self._final_pos = value

    def on_press(self):
        if self.state == 'down':
            self.selection_animation.start(self)
            self.rev_screen.market_formulation = self.text
            self.desc_box.text = '\n'.join(['[b]{0}[/b]'.format(self.text), self.desc])


class ValuationWizardRevenueSelect(Screen):
    """The revenue stream selection screen for the valuation wizard."""
    iso = StringProperty('')
    market_formulation = StringProperty('')
    has_selection = BooleanProperty(False)

    def __init__(self, **kwargs):
        super(ValuationWizardRevenueSelect, self).__init__(**kwargs)

        WizardRevToggleButton.desc_box = self.item_desc
        WizardRevToggleButton.rev_screen = self

    def on_pre_enter(self):
        self.iso = self.manager.get_screen('iso_select').iso

    def on_enter(self):
        Clock.schedule_once(partial(fade_in_animation, self.content), 0)

        def _fly_in_buttons(button, *args):
            button_anim = Animation(transition='out_expo', duration=1, opacity=1)
            button_anim &= Animation(transition='out_circ', duration=1, x=button.final_pos[0], y=button.final_pos[1])
            button_anim.start(button)

        for ix, rev_button in enumerate(reversed(self.rev_select.children), start=1):
            # Record the final intended position for each button.
            if rev_button.x < self.width:
                rev_button.final_pos = (rev_button.x, rev_button.y)
                rev_button.x = self.width
                #rev_button.y = 0

            Clock.schedule_once(partial(_fly_in_buttons, rev_button), ix*ANIM_STAGGER)

    def on_leave(self):
        Animation.stop_all(self.content, 'opacity')
        self.content.opacity = 0

        for rev_button in self.rev_select.children:
            Animation.cancel_all(rev_button)
            rev_button.x = self.width
            #rev_button.y = 0
            rev_button.opacity = 0

    def on_iso(self, instance, value):
        self.item_desc.text = ''
        while len(self.rev_select.children) > 0:
            for widget in self.rev_select.children:
                if isinstance(widget, WizardRevToggleButton):
                    self.rev_select.remove_widget(widget)

        node = self.manager.get_screen('node_select').node['nodeid']
        data_manager = App.get_running_app().data_manager
        market_models_available = data_manager.get_valuation_revstreams(value, node)

        self.has_selection = False

        for name, attributes in market_models_available.items():
            rev_button = WizardRevToggleButton(text=name, name=attributes['market type'],
                                               desc=attributes['desc'], opacity=0)
            self.rev_select.add_widget(rev_button)

    def on_market_formulation(self, instance, value):
        logging.info('ValuationWizard: Market formulation changed to {0}.'.format(value))

        self.has_selection = True if \
            any([button.state == 'down' for button in self.rev_select.children]) \
            else False

    def on_has_selection(self, instance, value):
        # Enables the next button if at least one revenue stream is selected.
        self.next_button.disabled = not value

    def _next_screen(self):
        """Adds the device selection screen if it does not exist and changes screens to it."""
        if not self.manager.has_screen('data_select'):
            screen = ValuationWizardDataSelect(name='data_select')
            self.manager.add_widget(screen)

        self.manager.transition.duration = BASE_TRANSITION_DUR
        self.manager.transition.direction = 'left'
        self.manager.current = 'data_select'


class WizardDataTileButton(ToggleTileButton):
    """Tile button with toggle behavior for selecting which batch of historical data to run."""
    data_screen = None

    def __init__(self, name, historical_data, **kwargs):
        super(WizardDataTileButton, self).__init__(**kwargs)

        self._name = name
        self._historical_data = historical_data

    @property
    def name(self):
        """The name of the historical data set."""
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def historical_data(self):
        """List of dictionaries with keys 'month' and 'year' corresponding to historical data."""
        return self._historical_data

    @historical_data.setter
    def historical_data(self, value):
        self._historical_data = value

    def on_press(self):
        if self.state == 'down':
            self.selection_animation.start(self)
            self.data_screen.selected_data_set = self

            self.data_screen.data_desc.text = '\n'.join([calendar.month_name[int(opt['month'])]
                                                         + ' ' + opt['year'] for opt in self.historical_data])


class ValuationWizardDataSelect(Screen):
    iso = StringProperty('')
    has_selection = BooleanProperty(False)
    selected_data_set = ObjectProperty()
    selected_data = ListProperty()

    def __init__(self, **kwargs):
        super(ValuationWizardDataSelect, self).__init__(**kwargs)

        WizardDataTileButton.data_screen = self

    def on_pre_enter(self):
        self.iso = self.manager.get_screen('iso_select').iso

    def on_enter(self):
        Clock.schedule_once(partial(fade_in_animation, self.content), 0)

        self.data_desc.text = ''
        while len(self.data_select.children) > 0:
            for widget in self.data_select.children:
                if isinstance(widget, WizardDataTileButton):
                    self.data_select.remove_widget(widget)

        self.has_selection = False

        node = self.manager.get_screen('node_select').node['nodeid']
        market_formulation = self.manager.get_screen('rev_stream_select').market_formulation
        data_manager = App.get_running_app().data_manager
        HISTORICAL_DATA = data_manager.get_historical_datasets(self.iso, node, market_formulation)
        
        try:
            for (name, data_set) in HISTORICAL_DATA.items():
                button = WizardDataTileButton(name=name, text=name, historical_data=data_set)
                self.data_select.add_widget(button)
        except AttributeError:
            pass



    def on_leave(self):
        Animation.stop_all(self.content, 'opacity')
        self.content.opacity = 0

    def on_iso(self, instance, value):
        self.data_desc.text = ''
        # while len(self.data_select.children) > 0:
        #     for widget in self.data_select.children:
        #         if isinstance(widget, WizardDataTileButton):
        #             self.data_select.remove_widget(widget)

        self.has_selection = False

        # try:
        #     for (name, data_set) in HISTORICAL_DATA[value].items():
        #         button = WizardDataTileButton(name=name, text=name, historical_data=data_set)
        #         self.data_select.add_widget(button)
        # except AttributeError:
        #     pass

        # node = self.manager.get_screen('node_select').node['nodeid']
        # market_formulation = self.manager.get_screen('rev_stream_select').market_formulation
        # data_manager = App.get_running_app().data_manager
        # HISTORICAL_DATA = data_manager.get_historical_datasets(value, node, market_formulation)
        
        # try:
        #     for (name, data_set) in HISTORICAL_DATA.items():
        #         button = WizardDataTileButton(name=name, text=name, historical_data=data_set)
        #         self.data_select.add_widget(button)
        # except AttributeError:
        #     pass

    def on_selected_data_set(self, instance, value):
        logging.info('ValuationWizard: Changed historical data set to {0}.'.format(value.name))

        self.selected_data = value.historical_data

        self.has_selection = True if \
            any([button.state == 'down' for button in self.data_select.children]) \
            else False

    def on_has_selection(self, instance, value):
        # Enables the next button if a data set is selected.
        self.next_button.disabled = not value

    def _next_screen(self):
        """Adds the pricing node selection screen if it does not exist and changes screens to it."""
        if not self.manager.has_screen('device_select'):
            screen = ValuationWizardDeviceSelect(name='device_select')
            self.manager.add_widget(screen)

        self.manager.transition.duration = BASE_TRANSITION_DUR
        self.manager.transition.direction = 'left'
        self.manager.current = 'device_select'


class WizardDeviceTileButton(ToggleTileButton):
    """Tile button with toggle behavior for each energy storage device option."""
    device_desc_box = None
    param_widget = None
    device_screen = None

    def __init__(self, desc, name, parameters, **kwargs):
        super(WizardDeviceTileButton, self).__init__(**kwargs)

        self._name = name
        self._desc = desc
        self._parameters = parameters

    @property
    def name(self):
        """The name of the energy storage device."""
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def desc(self):
        """A description of the energy storage device."""
        return self._desc

    @desc.setter
    def desc(self, value):
        self._desc = value

    @property
    def parameters(self):
        """Dictionary of the parameter names and values."""
        return self._parameters

    @parameters.setter
    def parameters(self, value):
        self._parameters = value

    def on_press(self):
        if self.state == 'down':
            self.selection_animation.start(self)

            self.device_desc_box.text = '\n'.join(['[b]{0}[/b]'.format(self.name), self.desc])
            self.param_widget.adjust_sliders(self)
            self.device_screen.device = self.name


class WizardDeviceParameterRow(GridLayout):
    """Grid layout containing parameter descriptor label, slider, and value label."""
    desc = DictProperty()

    def __init__(self, **kwargs):
        super(WizardDeviceParameterRow, self).__init__(**kwargs)

        self.name.text = self.desc['name']

        self.param_min = self.desc.get('min', 0)
        self.param_max = self.desc.get('max', 100)
        self.param_slider.min = self.param_min
        self.param_slider.max = self.param_max

        self.param_slider.value = self.desc.get('default', 50)
        self.param_slider.step = self.desc.get('step', 1)

    def _validate_input(self):
        """Validate entry when unfocusing text input."""
        if not self.text_input.focus:
            try:
                input_value = float(self.text_input.text)
            except ValueError:
                # No text entered.
                input_value = self.param_slider.value
                self.text_input.text = str(input_value)

                return

            if input_value > self.param_max or input_value < self.param_min:
                # If input value is out of range.
                popup = WarningPopup()
                popup.popup_text.text = '{param_name} must be between {param_min} and {param_max} (got {input_val}).'.format(param_name=self.name.text[:1].upper() + self.name.text[1:], param_min=self.param_min, param_max=self.param_max, input_val=input_value)
                popup.open()

                input_value = self.param_slider.value
                self.text_input.text = str(input_value)
            else:
                # Set slider value to input value.
                anim = Animation(transition='out_expo', duration=SLIDER_DUR, value=input_value)
                anim.start(self.param_slider)


class WizardDeviceParameterWidget(GridLayout):
    """Grid layout containing rows of parameter adjustment widgets."""
    def __init__(self, **kwargs):
        super(WizardDeviceParameterWidget, self).__init__(**kwargs)

        # build the widget by creating a row for each dictionary in param_list
        data_manager = App.get_running_app().data_manager
        PARAM_LIST = data_manager.get_valuation_wizard_device_params()

        for param in PARAM_LIST:
            row = WizardDeviceParameterRow(desc=param)
            self.add_widget(row)
            setattr(self, param['attr name'], row)

    def adjust_sliders(self, device):
        """Adjusts the parameter slider values based on the device selected."""
        for (param, val) in device.parameters.items():
            anim = Animation(transition='out_sine', duration=SLIDER_DUR, value=val)
            anim.start(getattr(self, param).param_slider)


class ParamTextInput(TextInput):
    """A TextInput field for entering parameter values."""

    def insert_text(self, substring, from_undo=False):
        # limit # chars
        substring = substring[:4 - len(self.text)]
        return super(ParamTextInput, self).insert_text(substring, from_undo=from_undo)


class ValuationWizardDeviceSelect(Screen):
    device = StringProperty('')
    has_selection = BooleanProperty(False)

    """The screen for describing the energy storage device for the valuation wizard."""
    def __init__(self, **kwargs):
        super(ValuationWizardDeviceSelect, self).__init__(**kwargs)

        WizardDeviceTileButton.device_screen = self
        WizardDeviceTileButton.device_desc_box = self.device_desc
        WizardDeviceTileButton.param_widget = self.param_widget

    def on_enter(self):
        Clock.schedule_once(partial(fade_in_animation, self.content), 0)

        def _fade_in_buttons(button, *args):
            button_anim = Animation(transition='out_expo', duration=FADEIN_DUR, opacity=1)
            button_anim.start(button)
        
        data_manager = App.get_running_app().data_manager
        DEVICE_LIST = data_manager.get_valuation_device_templates()

        if not self.device_select.children:
            for ix, dev in enumerate(DEVICE_LIST):
                device_button = WizardDeviceTileButton(text=dev['name'], name=dev['name'],
                                                       desc=dev['desc'],
                                                       parameters=dev['parameters'], opacity=0)
                self.device_select.add_widget(device_button)

        for ix, dev in enumerate(reversed(self.device_select.children)):
            Clock.schedule_once(partial(_fade_in_buttons, dev), ix*ANIM_STAGGER)

    def on_leave(self):
        Animation.stop_all(self.content, 'opacity')
        self.content.opacity = 0

        for button in self.device_select.children:
            Animation.cancel_all(button)
            button.opacity = 0

    def on_device(self, instance, value):
        logging.info('ValuationWizard: Energy storage device changed to {0}.'.format(value))

        self.has_selection = True if \
            any([button.state == 'down' for button in self.device_select.children]) \
            else False

    def on_has_selection(self, instance, value):
        # Enables the next button if a device is selected.
        self.next_button.disabled = not value

        # Enable parameter sliders and text inputs.
        for param_row in self.param_widget.children:
            param_row.text_input.disabled = False
            param_row.param_slider.disabled = False

    def _next_screen(self):
        """Adds the data selection screen if it does not exist and changes screens to it."""
        if not self.manager.has_screen('selection_summary'):
            screen = ValuationWizardSelectionSummary(name='selection_summary')
            self.manager.add_widget(screen)

        self.manager.transition.duration = BASE_TRANSITION_DUR
        self.manager.transition.direction = 'left'
        self.manager.current = 'selection_summary'


class WizardSummaryBubble(BoxLayout):
    pass


class ImageButton(ButtonBehavior, Image):
    pass


class WizardHeaderDesc():
    pass

# TODO: Check naming/style conventions: lower case and underscore delimited for class methods, camelcase for class names, etc.
# TODO: Use class rules for common widgets in kv. Example: The labels in the summary screen are all stylistically similar, so they should have a common class rule. (DRY: Don't Repeat Yourself)
# TODO: The WizardBodyText class (defined at top of kv file) is used for body text elements (as opposed to title elements) throughout the wizard. It inherits from the BodyTextBase (defined in look_and_feel.kv) which is the base class for body text throughout the application. This inheritance structure is useful to make sure the look and feel is consistent throughout. For example, if I wanted to change the default font of body text everywhere in the application, I would just change it in the BodyTextBase and it would propagate through the inheritance hierarchy.


class AboutSelectionSummary(ModalView):
    ISOselname = ''
    def __init__(self, **kwargs):
        super(AboutSelectionSummary, self).__init__(**kwargs)

    def openwebbrowser(self):
        if self.ISOselname  == "NYISO":
            webbrowser.open('http://www.nyiso.com/')
        elif self.ISOselname  == "ISONE":
            webbrowser.open('https://www.iso-ne.com/')
        elif self.ISOselname == "PJM":
            webbrowser.open('http://www.pjm.com/')
        elif self.ISOselname == "ERCOT":
            webbrowser.open('http://www.ercot.com/')
        elif self.ISOselname == "MISO":
            webbrowser.open('https://www.misoenergy.org/')
        elif self.ISOselname == "SPP":
            webbrowser.open('https://www.spp.org/')
        elif self.ISOselname == "CAISO":
            webbrowser.open('http://www.caiso.com/')
        else:
            webbrowser.open('http://www.sandia.gov/')

    def setisoinfo(self):
        # TODO: Make better descriptions, these are temporary taken from Wikipedia
        # print('Setting image')
        self.title_abt_selsum.text = '[ref=iso_title]{iso}[/ref]'.format(iso=self.ISOselname)

        if self.ISOselname == "NYISO":
            # self.img_abt_selsum.source = os.path.join('es_gui', 'resources', 'images', 'xNYISOlogo.jpg')
            self.lab_abt_selsum.text = "The New York Independent System Operator (NYISO) operates competitive wholesale markets to manage the flow of electricity across New York."
        elif self.ISOselname == "ISONE":
            # self.img_abt_selsum.source = os.path.join('es_gui', 'resources', 'images', 'xISO-NElogo.jpg')
            self.lab_abt_selsum.text = "ISO New England Inc. (ISO-NE) is an independent, non-profit Regional Transmission Organization (RTO), headquartered in Holyoke, Massachusetts, serving Connecticut, Maine, Massachusetts, New Hampshire, Rhode Island, and Vermont. ISO-NE oversees the operation of New England's bulk electric power system and transmission lines, generated and transmitted by its member utilities, as well as Hydro-Quebec, NB Power, the New York Power Authority and utilities in New York state, when the need arises. ISO-NE is responsible for reliably operating New England's bulk electric power generation and transmission system."
        elif self.ISOselname == "PJM":
            # self.img_abt_selsum.source = os.path.join('es_gui', 'resources', 'images', 'xPJMlogo.png')
            self.lab_abt_selsum.text = "PJM Interconnection LLC (PJM) is a regional transmission organization (RTO) in the United States. It is part of the Eastern Interconnection grid operating an electric transmission system serving all or parts of Delaware, Illinois, Indiana, Kentucky, Maryland, Michigan, New Jersey, North Carolina, Ohio, Pennsylvania, Tennessee, Virginia, West Virginia, and the District of Columbia."
        elif self.ISOselname == "ERCOT":
            # self.img_abt_selsum.source = os.path.join('es_gui', 'resources', 'images', 'xERCOTlogo.png')
            self.lab_abt_selsum.text = "The Electric Reliability Council of Texas (ERCOT) manages the flow of electric power on the Texas Interconnection that supplies power to 24 million Texas customers - representing 85 percent of the state's electric load."
        elif self.ISOselname == "MISO":
            # self.img_abt_selsum.source = os.path.join('es_gui', 'resources', 'images', 'xMISOlogo.png')
            self.lab_abt_selsum.text = 'The Midcontinent Independent System Operator, Inc., formerly named Midwest Independent Transmission System Operator, Inc. (MISO) is an Independent System Operator (ISO) and Regional Transmission Organization (RTO) providing open-access transmission service and monitoring the high-voltage transmission system in the Midwest United States and Manitoba, Canada and a southern United States region which includes much of Arkansas, Mississippi, and Louisiana.'
        elif self.ISOselname == "SPP":
            # self.img_abt_selsum.source = os.path.join('es_gui', 'resources', 'images', 'xMISOlogo.png')
            self.lab_abt_selsum.text = 'SPP description TBW.'
        elif self.ISOselname == "CAISO":
            # self.img_abt_selsum.source = os.path.join('es_gui', 'resources', 'images', 'xMISOlogo.png')
            self.lab_abt_selsum.text = 'CAISO description TBW.'

        else:
            # self.img_abt_selsum.source = os.path.join('es_gui', 'resources', 'images', 'SNLlogo.png')
            pass


class ValuationWizardSelectionSummary(Screen):
    """The selection summary screen for the valuation wizard."""
    iso = StringProperty('')
    has_selections = BooleanProperty(False)
    wiz_selections = DictProperty()

    # def on_iso(self, instance, value):
    #     self.node_desc.text = ''
    #
    #     # build options based on ISO selected
    #     try:
    #         ls = self.manager.parent.parent.manager.get_screen('load_data')
    #
    #         node_options = [{'name': ls.dms.get_node_name(x, self.iso),
    #                          'nodeid': x} for x in ls.nodeid[self.iso]]
    #         self.node_rv.data = node_options
    #     except AttributeError:
    #         pass

    def on_enter(self):
        Clock.schedule_once(partial(fade_in_animation, self.content), 0)

        # clear widgets
        # while len(self.content.children) > 0:
        #     for widget in self.content.children:
        #         if isinstance(widget, WizardSummaryBubble):
        #             self.content.remove_widget(widget)

        # extract selections from each screen
        summary_bubble = WizardSummaryBubble(size_hint=(1, 1))

        selections = {}
        selections['iso'] = self.manager.get_screen('rev_stream_select').iso
        selections['node'] = self.manager.get_screen('node_select').node['name']
        selections['nodeid'] = self.manager.get_screen('node_select').node['nodeid']
        selections['rev_streams'] = self.manager.get_screen('rev_stream_select').market_formulation
        selections['selected_data'] = self.manager.get_screen('data_select').selected_data
        selections['device'] = [x for x in self.manager.get_screen('device_select').param_widget.children]

        self.wiz_selections = selections  # assign to DictProperty of screen

        # convert to displayable text
        iso = self.wiz_selections['iso']
        node = self.wiz_selections['node']
        nodeid = self.wiz_selections['nodeid']
        rev_streams = self.wiz_selections['rev_streams']
        #hist_data = ', '.join([opt['name'] for opt in self.wiz_selections['selected_data']])
        device = '   \n'.join([
                     '   {0}: {1}'.format(x.name.text, str(x.param_slider.value))
                     for x in self.wiz_selections['device']
                 ])

        # summary_bubble.body.text += '\n\n'.join(
        #     [iso,
        #      node,
        #      rev_streams,
        #      hist_data,
        #      device,
        #      ]
        # )

        buttons_dev_sel = self.manager.get_screen('device_select').device_select.children
        but_press_dev_sel = [x.text for x in buttons_dev_sel if x.state == "down"]

        if not but_press_dev_sel:
            # print("empty list")
            but_press_dev_sel = "-"
        else:
            but_press_dev_sel = but_press_dev_sel[0]# print("NOT empty list")

        #print(but_press_dev_sel[0])

        isotextSW = '[u][ref={ref_in}]{textisosel} [/ref][/u]'.format(ref_in=iso, textisosel=iso)
        month_ini = calendar.month_name[int(self.wiz_selections['selected_data'][0]['month'])]
        year_ini = self.wiz_selections['selected_data'][0]['year']
        month_end = calendar.month_name[int(self.wiz_selections['selected_data'][-1]['month'])]
        year_end = self.wiz_selections['selected_data'][-1]['year']

        self.isolabel.text = '[b]Market Area:[/b] ' + isotextSW
        self.isolabel.markup = True
        self.isolabel.bind(on_ref_press=self.OpenAbSelSum)

        self.pcnodelabel.text = "[b]Pricing Node:[/b] " + node + " (id: " + str(nodeid) + ")"
        self.pcnodelabel.markup = True

        histdattext = '{m_i} {y_i} to {m_e} {y_e}'.format(m_i=month_ini, y_i=year_ini, m_e=month_end, y_e=year_end)
        self.histdatlabel.text = "[b]Dates Analyzed: [/b]" + histdattext
        self.histdatlabel.markup = True

        self.revstreamlabel.text = "[b]Revenue Streams: [/b]" + rev_streams
        self.revstreamlabel.markup = True

        self.esdevlabel.text = "[b]ES Device: [/b]" + but_press_dev_sel + '\n' + device
        self.esdevlabel.markup = True

        # print(rev_streams)
        # print(hist_data)
        # self.content.add_widget(summary_bubble)
        #
        #
        # # build bubbles for each selection made
        # iso_bubble = WizardSummaryBubble()
        # iso_bubble.cat_title.text = 'ISO'
        # iso_bubble.body.text = self.manager.get_screen('rev_stream_select').iso
        # self.content.add_widget(iso_bubble)
        #
        # rev_bubble = WizardSummaryBubble()
        # rev_bubble.cat_title.text = 'Revenue streams'
        # rev_selected = [opt.name for opt in self.manager.get_screen('rev_stream_select').rev_select.children
        #                 if opt.state == 'down']
        # rev_bubble.body.text = '\n'.join(rev_selected)
        # self.content.add_widget(rev_bubble)
        #
        # data_bubble = WizardSummaryBubble()
        # data_bubble.cat_title.text = 'Data'
        # selected_data = [x.selected_data for x in self.manager.get_screen('data_select').data_select.children
        #                  if x.state == 'down'][0]
        # data_bubble.body.text = '\n'.join([opt['name'] for opt in selected_data])
        # self.content.add_widget(data_bubble)
        #
        # node_bubble = WizardSummaryBubble()
        # node_bubble.cat_title.text = 'Node'
        # node_bubble.body.text = [x['name'] for x in self.manager.get_screen('node_select').node_rv.data
        #                          if x.get('selected', False)][0]
        # self.content.add_widget(node_bubble)
        #
        # device_bubble = WizardSummaryBubble()
        # device_bubble.cat_title.text = 'Device description'
        # device_bubble.body.text = '\n'.join([
        #     '{0}: {1}'.format(x.name.text, str(x.param_slider.value))
        #     for x in self.manager.get_screen('device_select').param_widget.children
        # ])
        # self.content.add_widget(device_bubble)

    def OpenAbSelSum(self,*args):
        # print(args[1])

        isoAboutSelectionSummary = AboutSelectionSummary()
        isoAboutSelectionSummary.ISOselname = args[1]
        isoAboutSelectionSummary.setisoinfo()
        isoAboutSelectionSummary.open()


    def on_leave(self):
        Animation.stop_all(self.content, 'opacity')
        self.content.opacity = 0

    def _next_screen(self):
        """Adds the execute screen if it does not exist and changes screens to it."""
        if not self.manager.has_screen('execute'):
            screen = ValuationWizardExecute(name='execute')
            self.manager.add_widget(screen)

        self.manager.transition.duration = BASE_TRANSITION_DUR
        self.manager.transition.direction = 'left'
        self.manager.current = 'execute'


class ValuationWizardExecute(Screen):
    """The screen for executing the prescribed optimizations for the valuation wizard."""
    solved_ops = []
    report_attributes = {}

    def on_enter(self):
        self.execute_run()

    def on_leave(self):
        self.progress_label.text = 'This may take a while. Please wait patiently!'

    def execute_run(self):
        valuation_home = self.manager.parent.parent.manager.get_screen('valuation_home')
        ss = self.manager.get_screen('selection_summary')

        # Parse selection summary for arguments to pass to ValOp handler.
        wiz_selections = ss.wiz_selections
        iso = wiz_selections['iso']
        node = wiz_selections['node']
        nodeid = wiz_selections['nodeid']

        data_manager = App.get_running_app().data_manager
        rev_streams = data_manager.get_valuation_revstreams(iso, nodeid)[wiz_selections['rev_streams']]['market type']

        hist_data = wiz_selections['selected_data']
        device = wiz_selections['device']

        handler_requests = {}
        handler_requests['iso'] = iso
        handler_requests['market type'] = rev_streams  # need this as string, not list
        handler_requests['months'] = [(entry['month'], entry['year']) for entry in hist_data]
        handler_requests['node id'] = nodeid
        handler_requests['param set'] = [{param.desc['attr name']: param.param_slider.value
                                         for param in device}]

        # Send requests to ValOp handler.
        valop_handler = valuation_home.handler
        valop_handler.solver_name = App.get_running_app().config.get('optimization', 'solver')
        self.solved_ops, handler_status = valop_handler.process_requests(handler_requests)

        popup = WizardCompletePopup()

        # Check ValOp handler status.
        if len(handler_status) > 0:
            if self.solved_ops:
                # At least one model solved successfully.
                popup.title = "Success!*"
                popup.popup_text.text = '\n'.join([
                    'All finished, but we found these issues:',
                ]
                + list(handler_status)
                )
            else:
                # No models solved successfully.
                popup.title = "Hmm..."
                popup.popup_text.text = '\n'.join([
                    'Unfortunately, none of the models were able to be solved. We found these issues:',
                ]
                + list(handler_status)
                )

                popup.results_button.text = "Take me back"
                popup.bind(on_dismiss=lambda x: self.manager.parent.parent.manager.nav_bar.go_up_screen())  # Go back to Valuation Home
                popup.open()
                return

        popup.bind(on_dismiss=self._next_screen)
        popup.open()

        # Save selection summary details to pass to report generator.
        deviceSelectionButtons = self.manager.get_screen('device_select').device_select.children
        selectedDeviceName = [x.text for x in deviceSelectionButtons if x.state == "down"][0]

        self.report_attributes = {'market area': iso,
                                  'pricing node': node,
                                  'selected device': selectedDeviceName,
                                  'dates analyzed': ' to '.join([
                                      ' '.join([calendar.month_name[int(hist_data[0]['month'])], hist_data[0]['year']]),
                                      ' '.join([calendar.month_name[int(hist_data[-1]['month'])], hist_data[-1]['year']]),
                                      ]),
                                  'revenue streams': wiz_selections['rev_streams'],
                                  'market type': rev_streams,
                                 }

        for param in device:
            self.report_attributes[param.desc['attr name']] = param.param_slider.value

    def _next_screen(self, *args):
        """Adds the report screen if it does not exist and changes screens to it."""
        report = ValuationReport(name='report', chart_data=self.solved_ops, market=self.report_attributes['market type'], report_attributes=self.report_attributes)
        self.manager.switch_to(report, direction='left', duration=BASE_TRANSITION_DUR)




from __future__ import absolute_import

import os
import pandas as pd
import webbrowser
from es_gui.apps.tech_selection import fAux

from kivy.app import App
from kivy.uix.screenmanager import Screen

from es_gui.resources.widgets.common import BASE_TRANSITION_DUR, WarningPopup, stnd_font
from es_gui.apps.tech_selection import fPlots


class TechSelectionFeasible(Screen):
    """"""

    def __init__(self, **kwargs):
        super(TechSelectionFeasible, self).__init__(**kwargs)

    def on_pre_enter(self):
        """Clear any widgets already present in the screen and create a new widget displaying the user selections."""

        # Clear all previous widgets
        if self.param_widget.children:
            self.param_widget.clear_widgets()

        # Rebuild the widget displaying a summary of user selections
        self.param_widget.build()
        for child in self.param_widget.children:
            child.name.font_size, child.text_input.font_size = stnd_font, stnd_font
            child.name.size_hint_x, child.text_input.size_hint_x = 0.35, 0.65

        # Reload the heatmap of feasible technologies
        self.plot_feasibility_image.reload()

    def _next_screen(self):
        if not self.manager.has_screen('ranking_techs'):
            screen = TechSelectionRanking(name='ranking_techs')
            self.manager.add_widget(screen)

        self.manager.transition.duration = BASE_TRANSITION_DUR/2
        self.manager.transition.direction = 'left'
        self.manager.current = 'ranking_techs'


class TechSelectionRanking(Screen):
    """"""
    def __init__(self, **kwargs):
        super(TechSelectionRanking, self).__init__(**kwargs)

    def on_pre_enter(self):
        """Clear any widgets already present in the screen and create a new widget displaying the user selections."""

        # Collect user selections values from a .json file
        data_manager = App.get_running_app().data_manager
        MODEL_PARAMS = data_manager.get_tech_selection_params()
        user_selections = {val['attr name']: val['value'] for val in MODEL_PARAMS}

        self.target_cost_kW = 1500
        self.target_cost_kWh = 1000

        if user_selections['app_type'] == 'Energy':
            self.togglebutton_kWh.state = 'down'
            self.togglebutton_kW.state = 'normal'
            self.target_cost_value.text = str(self.target_cost_kWh)
        elif user_selections['app_type'] == 'Power':
            self.togglebutton_kWh.state = 'normal'
            self.togglebutton_kW.state = 'down'
            self.target_cost_value.text = str(self.target_cost_kW)
        else:
            pass

        if self.param_widget.children:
            self.param_widget.clear_widgets()
        self.param_widget.build()

        for child in self.param_widget.children:
            child.name.font_size, child.text_input.font_size = stnd_font, stnd_font
            child.name.size_hint_x, child.text_input.size_hint_x = 0.35, 0.65

        self.plot_ranking_image.reload()

    def help_weights(self):
        """Create popup window for showing help about the weights for each factor in the feasibility computation."""

        # Text for the popup window
        popup_txt = ('The total feasibility score for each energy storage technology is given as the weighted geometric '
                     'mean of the [i]application[/i], [i]location[/i], and [i]cost[/i] scores, which is then multiplied by '
                     'the [i]maturity[/i] score. This approach guarantees that only market-ready technologies receive a '
                     'high total feasibility score.\n\nBy default, all factors have the same weight. However, the user can '
                     'modify the weights to prioritize a subset of these factors.\n\nSetting a weight to 0.0 effectively '
                     'removes the corresponding factor from the final computation.')

        # Create, populate, and display the popup window
        popup = WarningPopup(size_hint=(0.5, 0.52))
        popup.dismiss_button.size_hint_y = 0.05/0.52
        popup.title = 'HELP: weights for each factor in the total feasibility score computation'
        popup.popup_text.text = popup_txt
        popup.popup_text.markup = True
        popup.open()

    def help_target_cost(self):
        """Create popup window for showing help about the target cost for the cost score computation."""

        # Text for the popup window
        popup_txt = ('The cost score for each energy storage technology is inversely proportional to its total capital cost '
                     'and normalized with respect to a desired target cost; i.e.:\n\n[color=ffa300][i]Cost score = Target '
                     'cost / (Capital cost + Target cost)[/i][/color]\n\nNote that the cost score at the desired target '
                     'cost is 0.5.\n\nThe target cost is usually given in $/kW for power applications and in $/kWh for '
                     'energy applications.')

        # Create, populate, and display the popup window
        popup = WarningPopup(size_hint=(0.5, 0.5))
        popup.dismiss_button.size_hint_y = 0.05/0.5
        popup.title = 'HELP: target cost for the cost score computation'
        popup.popup_text.text = popup_txt
        popup.popup_text.markup = True
        popup.open()

    def update_scores(self):
        """"""

        # TODO: read this file only if it hasn't been read previously
        self.dfRanking = pd.read_csv(os.path.join('results', 'tech_selection', 'table_ranking.csv'), index_col=0)

        test_aux = pd.DataFrame(self.dfRanking[['Application score', 'Location score', 'Cost score', 'Maturity score']])

        xx = test_aux['Cost score'].values
        old_target = 1500
        new_target = float(self.target_cost_value.text)
        #bb = new_target / (old_target*(1-xx)/xx + new_target)
        bb = new_target*xx / (old_target*(1-xx) + new_target*xx) # Avoid "division by zero" warnings
        test_aux['Cost score'] = bb
        self.dfRanking['Cost score'] = bb

        self.dfRanking['Total score'] = test_aux['Maturity score'] *\
            fAux.geom_mean(test_aux[['Application score', 'Location score', 'Cost score']],
                           weights=[self.app_slider.value, self.location_slider.value, self.cost_slider.value])
        self.dfRanking.fillna(value=0, inplace=True)
        self.dfRanking.sort_values(by=['Total score', 'Application score', 'Location score', 'Cost score', 'Maturity score'],
                                   inplace=True)

        # Plot: final feasibility scores
        fig = fPlots.plot_ranking_techs(self.dfRanking)
        fig.savefig(os.path.join('results', 'tech_selection', 'plot_ranking.png'))

        self.plot_ranking_image.reload()

        type_of_cost = [tb.text for tb in [self.togglebutton_kW, self.togglebutton_kWh] if tb.state == 'down'][0]

        popup = WarningPopup(size_hint=(0.4, 0.35))
        popup.dismiss_button.size_hint_y = 0.05/0.35
        popup.title = 'Success!'
        popup.popup_text.text = (
            f'The total feasibility score has been recomputed with the following settings:\n\n'
            f'- Weights: {self.app_slider.value:.1f} (application), {self.location_slider.value:.1f} (location), and '
            f'{self.cost_slider.value:.1f} (cost).\n'
            f'- Target cost: {self.target_cost_value.text} {type_of_cost}.'
            )
        popup.open()

    def _next_screen(self):
        if not self.manager.has_screen('save_results_techs'):
            screen = TechSelectionSaveResults(name='save_results_techs')
            self.manager.add_widget(screen)

        self.manager.transition.duration = BASE_TRANSITION_DUR/2
        self.manager.transition.direction = 'left'
        self.manager.current = 'save_results_techs'


class TechSelectionSaveResults(Screen):
    """"""
    def __init__(self, **kwargs):
        super(TechSelectionSaveResults, self).__init__(**kwargs)
        # self.sm = sm

    def feasibility_export_png(self):
        """Export currently displayed figure to .png file in specified location."""
        filename = self.name_plot_feasibility.text
        if filename == '':
            filename = 'out_plot_feasibility'
        outname = os.path.join('results', 'tech_selection', f'{filename}.png')
        self.manager.get_screen('feasible_techs').plot_feasibility.children[0].export_to_png(outname)

        popup = WarningPopup(size_hint=(0.4, 0.4))
        popup.title = 'Success!'
        popup.popup_text.text = f'Figure successfully exported to:\n\n {os.path.abspath(outname)}'
        popup.open()

    def ranking_export_png(self):
        """Export currently displayed figure to .png file in specified location."""
        filename = self.name_plot_ranking.text
        if filename == '':
            filename = 'out_plot_ranking'
        outname = os.path.join('results', 'tech_selection', f'{filename}.png')
        self.manager.get_screen('ranking_techs').plot_ranking.children[0].export_to_png(outname)

        popup = WarningPopup(size_hint=(0.4, 0.4))
        popup.title = 'Success!'
        popup.popup_text.text = f'Figure successfully exported to:\n\n {os.path.abspath(outname)}'
        popup.open()

    def ranking_export_csv(self):
        """Export ranking DataFrame to .csv file in specified location."""
        filename = self.name_table_ranking.text
        if filename == '':
            filename = 'out_table_ranking'
        outname = os.path.join('results', 'tech_selection', f'{filename}.csv')
        self.manager.get_screen('ranking_techs').dfRanking.to_csv(outname)

        popup = WarningPopup(size_hint=(0.4, 0.4))
        popup.title = 'Success!'
        popup.popup_text.text = 'Results file successfully exported to:\n\n' + os.path.abspath(outname)
        popup.open()


# -*- coding: utf-8 -*-
"""
Created on Tue Nov 13 14:48:42 2018

@author: tunguy
"""
import urllib.request
import os
import pandas as pd
from pandas.io.json import json_normalize
import numpy as np
import json
import holidays
from datetime import datetime
import datetime as dt

def download_utdata(**kwargs):
    url_iou="https://openei.org/doe-opendata/dataset/53490bd4-671d-416d-aae2-de844d2d2738/resource/500990ae-ada2-4791-9206-01dc68e36f12/download/iouzipcodes2017.csv"
    url_noniou="https://openei.org/doe-opendata/dataset/53490bd4-671d-416d-aae2-de844d2d2738/resource/672523aa-0d8a-4e6c-8a10-67e311bb1691/download/noniouzipcodes2017.csv"
    
    proxy_dict={}
    http_proxy_on=True
    https_proxy_on=True
    succeed=True
    try:
        proxy_dict["http"]=kwargs["http_proxy"]
    except:
        http_proxy_on=False

    try:
        proxy_dict["https"]=kwargs["https_proxy"]
    except:
        https_proxy_on=False
    try:
        des_dir = kwargs["dirloc"]
    except:
        des_dir="./"    
        
    if not os.path.exists(des_dir):
        os.makedirs(des_dir)
        
    des_file_iou = des_dir+"iouzipcodes2017.csv"
    des_file_noniou = des_dir+"noniouzipcodes2017.csv"
    proxy_support = urllib.request.ProxyHandler(proxy_dict)
    opener = urllib.request.build_opener(proxy_support)
    urllib.request.install_opener(opener)
    try:
        urllib.request.urlretrieve(url_iou, des_file_iou)  
    except urllib.error.URLError as e1:
        print(e1)
        succeed=False
    try:
        urllib.request.urlretrieve(url_noniou, des_file_noniou)  
    except urllib.error.URLError as e2:
        print(e2)
        suceed=False
        
    if succeed:
        df_iou = pd.read_csv(des_file_iou)
        df_noniou= pd.read_csv(des_file_noniou)
        utdataframe=pd.concat([df_iou,df_noniou],ignore_index=True)
    else:
        utdataframe=pd.DataFrame()
    
    return utdataframe

def search_utdata_byname(**kwargs):
    
    try:
        utdataframe=kwargs["utdataframe"]
    except:
        utdataframe=pd.DataFrame()
            
    try:
        utname=kwargs["utname"]
    except:
        utname="na"
           
    if len(utdataframe)>0:
        utdatabyname=utdataframe.loc[utdataframe['utility_name']==utname]
        utdatabyname=utdatabyname[['eiaid','utility_name','state','ownership']]
        utdatabyname=utdatabyname.drop_duplicates()
    else:
        utdatabyname=pd.Dataframe()
    
    return utdatabyname

def search_utdata_byzip(**kwargs):
    
    try:
        utdataframe=kwargs["utdataframe"]
    except:
        utdataframe=pd.DataFrame()
            
    try:
        utzip=kwargs["utzip"]
    except:
        utzip="na"
           
    if len(utdataframe)>0:
        utdatabyzip=utdataframe.loc[utdataframe['zip']==utzip]
        utdatabyzip=utdatabyzip[['eiaid','utility_name','state','ownership']]
        utdatabyzip=utdatabyzip.drop_duplicates()
    else:
        utdatabyzip=pd.Dataframe()
    
    return utdatabyzip

def search_utdata_bystate(**kwargs):
    
    try:
        utdataframe=kwargs["utdataframe"]
    except:
        utdataframe=pd.DataFrame()
            
    try:
        utstate=kwargs["utstate"]
    except:
        utstate="na"
           
    if len(utdataframe)>0:
        utdatabystate=utdataframe.loc[utdataframe['state']==utstate]
        utdatabystate=utdatabystate[['eiaid','utility_name','state','ownership']]
        utdatabystate=utdatabystate.drop_duplicates()
    else:
        utdatabystate=pd.Dataframe()
    
    return utdatabystate

def find_utschld(**kwargs):
    proxy_dict={}
    http_proxy_on=True
    https_proxy_on=True
    succeed=True
    
    try:
        proxy_dict["http"]=kwargs["http_proxy"]
    except:
        http_proxy_on=False

    try:
        proxy_dict["https"]=kwargs["https_proxy"]
    except:
        https_proxy_on=False
    
    try:
        eia_id="&eia="+kwargs["eiaid"]
    except:
        eia_id="no_id"
    
    try:
        api_key="&api_key="+kwargs["api_key"]
    except:
        api_key="no_key"
       
    root_url="https://api.openei.org/utility_rates?"
    version="version=5"
    request_format="&format=json"
    detail="&detail=full"
    url_openei=root_url+version+request_format+api_key+eia_id+detail
    print(url_openei)
    
    try:
        des_dir = kwargs["dirloc"]
    except:
        des_dir="./"    
    
    if not os.path.exists(des_dir):
        os.makedirs(des_dir)
    
    des_file=des_dir+"schedule.json"
    
    proxy_support = urllib.request.ProxyHandler(proxy_dict)
    opener = urllib.request.build_opener(proxy_support)
    urllib.request.install_opener(opener)
    try:
        urllib.request.urlretrieve(url_openei, des_file)  
    except urllib.error.URLError as e:
        print(e)
        succeed=False
    
    schld=json.loads(open(des_file).read()) 
#    schld_data=json_normalize(schld['items'])
    return schld['items'] # this is a list of schedules in which each element contains the full details of a schedule.

def read_load_profile(path, month):
    """Reads the annual load profile file located at path and returns the array of the load profile for the given month."""
    load_df = pd.read_csv(path)

    if isinstance(month, str):
        month = int(month)

    # Assumptions: column 0 is datetime, column 1 is data
    datetime_column_name = load_df.columns[0]
    data_column_name = load_df.columns[-1]

    # Overwrite DateTime column (esp. for data obtained from OpenEI)
    datetime_start = datetime(2019, 1, 1, 0)
    hour_range = pd.date_range(start=datetime_start, periods=len(load_df), freq="H")
    load_df[datetime_column_name] = hour_range

    # Filter by given month.
    datetime_column = pd.to_datetime(load_df[datetime_column_name])
    load_df_month = load_df.loc[datetime_column.apply(lambda x: x.month == month)]
    load_profile = load_df_month[data_column_name].values

    return load_profile

def read_pv_profile(path, month):
    """Reads the annual PV profile file located at path and returns the array of the PV profile for the given month."""
    if isinstance(month, str):
        month = int(month)
    
    with open(path) as f:
        profile_obj = json.load(f)
        
    pv_output_w = profile_obj['outputs']['ac']

    # Convert to kW.
    df_pv_output = pd.DataFrame(pv_output_w, columns=['kW'])*1e-3

    # Apply datetime index for filtering.
    datetime_start = datetime(2019, 1, 1, 0)
    hour_range = pd.date_range(start=datetime_start, periods=len(pv_output_w), freq='H')
    df_pv_output['dt'] = hour_range

    # Filter by given month.
    df_pv_month = df_pv_output.loc[df_pv_output['dt'].apply(lambda x: x.month == month)]
    pv_output_kw = df_pv_month['kW'].values

    return pv_output_kw

def get_pv_profile_string(path):
    """Reads the PV profile JSON object and returns a list of string descriptors."""
    with open(path) as f:
        profile_obj = json.load(f)
    
    module_type_list = ['Standard', 'Premium', 'Thin Film', 'N/A']
    array_type_list = ['Fixed (open rack)', 'Fixed (roof mounted)', '1-axis', '1-axis (backtracking)', '2-axis', 'N/A']
    
    query_inputs = profile_obj['inputs']

    # Skip if this is a custom/imported profile.
    if query_inputs["array_type"] == -1:
        return ["Custom",]

    coordinates = 'Location: {lat}, {lon}'.format(lat=query_inputs['lat'], lon=query_inputs['lon'])
    system_capacity = 'System Capacity: {0} kW'.format(query_inputs['system_capacity'])
    azimuth = 'Azimuth: {0} deg'.format(query_inputs['azimuth'])
    tilt = 'Tilt: {0} deg'.format(query_inputs['tilt'])
    array_type = 'Array Type: {0}'.format(array_type_list[int(query_inputs['array_type'])])
    module_type = 'Module Type: {0}'.format(module_type_list[int(query_inputs['module_type'])])
    system_losses = 'System Losses: {0}%'.format(query_inputs['losses'])

    descriptors = [
        coordinates,
        system_capacity,
        azimuth,
        tilt,
        array_type,
        module_type,
        system_losses,
    ]

    return descriptors


def input_df(year,wkday_eschld,wkend_eschld,wkday_dschld,wkend_dschld):
    """
    Generate a Pandas dataframe from the downloaded rate schedule data.
    :param year: an integer.
    :param wkday_eschld: a 12x24 NumPy array, wkday_schld[i][j] is the tou energy schedule of hour j+1 in a weekday of month i+1
    :param wkend_eschld: a 12x24 NumPy array, wkend_schld[i][j] is the tou energy schedule of hour j+1 in a weekend day or a holiday of month i+1
    :param wkday_dschld: a 12x24 NumPy array, wkday_schld[i][j] is the tou demand schedule of hour j+1 in a weekday of month i+1
    :param wkend_dschld: a 12x24 NumPy array, wkend_schld[i][j] is the tou demand schedule of hour j+1 in a weekend day or a holiday of month i+1
    
    :return opt_df: a Pandas dataframe with the following template:|year|month|day|hour|tou_energy_schedule|tou_demand_schedule|
    """
    column_list=['year','month','day','hour','tou_energy_schedule','tou_demand_schedule']
    data_list=[]
    nday={1:31,2:28,3:31,4:30,5:31,6:30,7:31,8:31,9:30,10:31,11:30,12:31}
    
    if int(year)%4==0:
        nday[2]=29
        
    us_holidays=holidays.UnitedStates()
    for month in range(1,13):
        for day in range(1,nday[month]+1):
            date=datetime(year, month, day)
            dayofweek = date.weekday()
            for hour in range(1,25):
                if (date in us_holidays) or (dayofweek>4):
                    if len(wkend_dschld)>1:  
                        row=[year,month,day,hour,wkend_eschld[month-1][hour-1],wkend_dschld[month-1][hour-1]]
                    else:
                        row=[year,month,day,hour,wkend_eschld[month-1][hour-1],0]
                else:
                    if len(wkday_dschld)>1:  
                        row=[year,month,day,hour,wkday_eschld[month-1][hour-1],wkday_dschld[month-1][hour-1]]
                    else:
                        row=[year,month,day,hour,wkday_eschld[month-1][hour-1],0]
                    
                data_list.append(row)
    opt_df=pd.DataFrame(data=data_list,columns=column_list)

    return opt_df
                    
# Example             
#utdataframe = download_utdata(https_proxy="wwwproxy.sandia.gov:80")
#utdata=search_utdata_byzip(utdataframe=utdataframe,utzip=87185)
#utid=utdata['eiaid'].values
#utid=[14328]
#print(utid)
#schld_data=find_utschld(https_proxy="wwwproxy.sandia.gov:80",eiaid=str(utid[0]),api_key="QqYsZ7wI57tBiNaqc4IZfffvgWaorWER6J3D69tw")
#
#for i in range(len(schld_data)):
#    print(i,schld_data[i]['name'])
#
#print(schld_data[471].keys())
#print(schld_data[471]['flatdemandstructure'])
#
#wkday_eschld=schld_data[471]['energyweekdayschedule']
#wkend_eschld=schld_data[471]['energyweekendschedule']
#tou_erate=schld_data[471]['energyratestructure']
#
#wkday_dschld=schld_data[471]['demandweekdayschedule']
#wkend_dschld=schld_data[471]['demandweekendschedule']
#tou_drate=schld_data[471]['demandratestructure']
#
#flat_dschld=schld_data[471]['flatdemandmonths']
#flat_drate=schld_data[471]['flatdemandstructure']
#
#year=2016
#opt_df=input_df(year,wkday_eschld,wkend_eschld,wkday_dschld,wkend_dschld)
#opt_df.to_csv('./input.csv')
#print(opt_df)


from __future__ import absolute_import
import math

from pyomo.environ import *


class ExpressionsBlock:
    """Creates blocks for objective and constraint functions and assigns them to the Pyomo model."""

    def __init__(self, market_type):
        self._market_type = market_type

    @property
    def market_type(self):
        """The market formulation to create blocks for."""
        return self._market_type

    @market_type.setter
    def market_type(self, value):
        self._market_type = value

    def set_expressions(self, model):
        """Generates the objective and constraint expressions for model."""
        block_obj = Block()
        model.objectives = block_obj

        block_con = Block()
        model.constraints = block_con

        if self.market_type == 'arbitrage':
            self._objective_arb(block_obj)
            self._constraints_arb(block_con)
        elif self.market_type == 'ercot_arbreg':
            self._objective_ercot_arbreg(block_obj)
            self._constraints_ercot_arbreg(block_con)
        elif self.market_type == 'pjm_pfp':
            self._objective_pjm_pfp(block_obj)
            self._constraints_pjm_pfp(block_con)
        elif self.market_type == 'miso_pfp':
            self._objective_miso_pfp(block_obj)
            self._constraints_miso_pfp(block_con)
        elif self.market_type == 'isone_pfp':
            self._objective_isone_pfp(block_obj)
            self._constraints_isone_pfp(block_con)
        ################################################################
        elif self.market_type == 'nyiso_pfp':
            self._objective_nyiso_pfp(block_obj)
            self._constraints_nyiso_pfp(block_con)
        ################################################################
        elif self.market_type == 'spp_pfp':
            self._objective_spp_pfp(block_obj)
            self._constraints_spp_pfp(block_con)
        ################################################################
        elif self.market_type == 'caiso_pfp':
            self._objective_caiso_pfp(block_obj)
            self._constraints_caiso_pfp(block_con)
        ################################################################
        else:
            raise ValueError('Invalid market type specified!')

    def _objective_arb(self, block):
        eq_objective_arb(block)

    def _constraints_arb(self, block):
        eq_stateofcharge_arb(block)
        eq_stateofcharge_initial(block)
        eq_stateofcharge_final(block)
        ineq_stateofcharge_minimum(block)
        ineq_stateofcharge_maximum(block)
        ineq_power_limit(block)
        #ineq_charge_limit(block)
        #ineq_discharge_limit(block)

    def _objective_ercot_arbreg(self, block):
        eq_objective_ercot_arbreg(block)

    def _constraints_ercot_arbreg(self, block):
        eq_stateofcharge_ercot_arbreg(block)
        eq_stateofcharge_initial(block)
        eq_stateofcharge_final(block)
        ineq_stateofcharge_minimum_reserve_twoprod(block)
        ineq_stateofcharge_maximum_reserve_twoprod(block)
        ineq_power_limit_twoprod(block)
        # ineq_charge_limit_ercot_arbreg(block)
        # ineq_discharge_limit_ercot_arbreg(block)

    def _objective_pjm_pfp(self, block):
        eq_objective_pjm_pfp(block)

    def _constraints_pjm_pfp(self, block):
        eq_stateofcharge_pjm_pfp(block)
        eq_stateofcharge_initial(block)
        eq_stateofcharge_final(block)
        ineq_stateofcharge_minimum_reserve_oneprod(block)
        ineq_stateofcharge_maximum_reserve_oneprod(block)
        ineq_power_limit_oneprod(block)
        # ineq_charge_limit_pfp(block)
        # ineq_discharge_limit_pfp(block)

    def _objective_miso_pfp(self, block):
        eq_objective_miso_pfp(block)

    def _constraints_miso_pfp(self, block):
        eq_stateofcharge_miso_pfp(block)
        eq_stateofcharge_initial(block)
        eq_stateofcharge_final(block)
        ineq_stateofcharge_minimum_reserve_oneprod(block)
        ineq_stateofcharge_maximum_reserve_oneprod(block)
        ineq_power_limit_oneprod(block)
        # ineq_charge_limit_pfp(block)
        # ineq_discharge_limit_pfp(block)

    def _objective_isone_pfp(self, block):
        eq_objective_isone_pfp(block)

    def _constraints_isone_pfp(self, block):
        eq_stateofcharge_isone_pfp(block)
        eq_stateofcharge_initial(block)
        eq_stateofcharge_final(block)
        ineq_stateofcharge_minimum_reserve_oneprod(block)
        ineq_stateofcharge_maximum_reserve_oneprod(block)
        ineq_power_limit_oneprod(block)
        # ineq_charge_limit_pfp(block)
        # ineq_discharge_limit_pfp(block)

    ####################################################################################################################
    # NYISO
    ####################################################################################################################
    def _objective_nyiso_pfp(self, block):
        eq_objective_nyiso_pfp(block)

    def _constraints_nyiso_pfp(self, block):
        eq_stateofcharge_nyiso_pfp(block)
        eq_stateofcharge_initial(block)
        eq_stateofcharge_final(block)
        ineq_stateofcharge_minimum_reserve_oneprod(block)
        ineq_stateofcharge_maximum_reserve_oneprod(block)
        ineq_power_limit_oneprod(block)
        # ineq_charge_limit_pfp(block)
        # ineq_discharge_limit_pfp(block)
    ####################################################################################################################
    ####################################################################################################################
    # SPP
    ####################################################################################################################
    def _objective_spp_pfp(self, block):
        eq_objective_spp_pfp(block)

    def _constraints_spp_pfp(self, block):
        eq_stateofcharge_spp_pfp(block)
        eq_stateofcharge_initial(block)
        eq_stateofcharge_final(block)
        ineq_stateofcharge_minimum_reserve_twoprod(block)
        ineq_stateofcharge_maximum_reserve_twoprod(block)
        ineq_power_limit_twoprod(block)
        # ineq_charge_limit_pfp(block)
        # ineq_discharge_limit_pfp(block)
    ####################################################################################################################
    ####################################################################################################################
    # CAISO
    ####################################################################################################################
    def _objective_caiso_pfp(self, block):
        eq_objective_caiso_pfp(block) ## TODO

    def _constraints_caiso_pfp(self, block):
        eq_stateofcharge_caiso_pfp(block) ## TODO
        eq_stateofcharge_initial(block)
        eq_stateofcharge_final(block)
        ineq_stateofcharge_minimum_reserve_twoprod(block)
        ineq_stateofcharge_maximum_reserve_twoprod(block)
        ineq_power_limit_twoprod(block)
        # ineq_charge_limit_pfp(block)
        # ineq_discharge_limit_pfp(block)
    ####################################################################################################################

#############################
# Arbitrage only ############
#############################


def eq_objective_arb(m):
    """Net revenue over the time horizon for arbitrage only."""
    mp = m.parent_block()

    _expr = sum(
        (mp.price_electricity[t] * mp.q_d[t] - mp.price_electricity[t] * mp.q_r[t]) * math.e ** (-t * mp.R) for
        t in mp.time)

    mp.objective_expr += _expr
    m.objective_rt = Expression(expr=_expr)


def eq_stateofcharge_arb(m):
    """Definition of state of charge for device in participating in arbitrage only."""
    mp = m.parent_block()

    def _eq_stateofcharge_arb(_m, t):
        return mp.Self_discharge_efficiency * mp.s[t] + mp.Round_trip_efficiency * mp.q_r[t] \
            - mp.q_d[t] == mp.s[t+1]       

    m.stateofcharge = Constraint(mp.time, rule=_eq_stateofcharge_arb)


def ineq_power_limit(m):
    """Limits the energy charged and discharged at each timestep to the device power rating."""
    mp = m.parent_block()

    def _ineq_power_limit(_m, t):
        return mp.Power_rating >= mp.q_r[t] + mp.q_d[t]

    m.power_limit = Constraint(mp.time, rule=_ineq_power_limit)


def ineq_charge_limit(m):
    """Probably deprecated. Limits the energy charged at each timestep to the device maximum."""
    mp = m.parent_block()

    def _ineq_charge_limit(_m, t):
        return mp.Q_r_max >= mp.q_r[t]

    m.charge_limit = Constraint(mp.time, rule=_ineq_charge_limit)


def ineq_discharge_limit(m):
    """Probably deprecated. Limits the energy discharged at each timestep to the device maximum."""
    mp = m.parent_block()

    def _ineq_discharge_limit(_m, t):
        return mp.Q_d_max >= mp.q_d[t]

    m.discharge_limit = Constraint(mp.time, rule=_ineq_discharge_limit)


###################################
# ERCOT Arbitrage and regulation ##
###################################


def eq_objective_ercot_arbreg(m):
    """Net revenue over the time horizon for ERCOT arbitrage and regulation."""

    mp = m.parent_block()

    _expr = sum((mp.price_electricity[t] * mp.q_d[t] - mp.price_electricity[t] * mp.q_r[t]
                 + mp.price_reg_up[t] * mp.q_ru[t] + mp.price_reg_down[t] * mp.q_rd[t]
                 + mp.price_electricity[t] * mp.q_ru[t] * mp.fraction_reg_up[t]
                 - mp.price_electricity[t] * mp.q_rd[t] * mp.fraction_reg_down[t])
                * math.e ** (-t * mp.R) for t in mp.time)

    mp.objective_expr += _expr
    m.objective_rt = Expression(expr=_expr)


def eq_stateofcharge_ercot_arbreg(m):
    """Definition of state of charge for device participating in ERCOT arbitrage and regulation."""
    mp = m.parent_block()

    def _eq_stateofcharge_ercot_arbreg(_m, t):
        return mp.Self_discharge_efficiency * mp.s[t] + mp.Round_trip_efficiency * mp.q_r[t] \
                - mp.q_d[t] + mp.Round_trip_efficiency * mp.fraction_reg_down[t] * mp.q_rd[t] \
                - mp.fraction_reg_up[t] * mp.q_ru[t] == mp.s[t+1]

    m.stateofcharge = Constraint(mp.time, rule=_eq_stateofcharge_ercot_arbreg)


def ineq_charge_limit_ercot_arbreg(m):
    """Limits the energy charged at each timestep to the device maximum."""
    mp = m.parent_block()

    def _ineq_charge_limit_arbreg(_m, t):
        return mp.Q_r_max >= mp.q_r[t] + mp.q_rd[t]

    m.charge_limit = Constraint(mp.time, rule=_ineq_charge_limit_arbreg)


def ineq_discharge_limit_ercot_arbreg(m):
    """Limits the energy discharged at each timestep to the device maximum."""
    mp = m.parent_block()

    def _ineq_discharge_limit_ercot_arbreg(_m, t):
        return mp.Q_d_max >= mp.q_d[t] + mp.q_ru[t]

    m.discharge_limit = Constraint(mp.time, rule=_ineq_discharge_limit_ercot_arbreg)


#############################
# PJM Pay-for-Performance ###
#############################


def eq_objective_pjm_pfp(m):
    """Net revenue over the time horizon for pay-for-performance in PJM market."""
    mp = m.parent_block()

    _expr = sum((mp.price_electricity[t] * mp.q_d[t] - mp.price_electricity[t] * mp.q_r[t]
                 + mp.q_reg[t] * mp.perf_score[t] * (mp.mi_mult[t] * mp.price_reg_service[t] + mp.price_regulation[t]))
                * math.e ** (-t * mp.R) for t in mp.time)

    mp.objective_expr += _expr
    m.objective_rt = Expression(expr=_expr)


def eq_stateofcharge_pjm_pfp(m):
    """Definition of state of charge for device in PJM pay-for-performance market."""
    mp = m.parent_block()

    def _eq_stateofcharge_pjm_pfp(_m, t):
        return mp.Self_discharge_efficiency * mp.s[t] + mp.Round_trip_efficiency * mp.q_r[t] \
                - mp.q_d[t] + mp.Round_trip_efficiency * mp.fraction_reg_down[t] * mp.q_reg[t] \
                - mp.fraction_reg_up[t] * mp.q_reg[t] == mp.s[t+1]

    m.stateofcharge = Constraint(mp.time, rule=_eq_stateofcharge_pjm_pfp)


def ineq_charge_limit_pjm_pfp(m):
    """Limits the energy charged at each timestep to the device maximum."""
    mp = m.parent_block()

    def _ineq_charge_limit_pjm_pfp(_m, t):
        return mp.Q_r_max >= mp.q_r[t] + mp.q_reg[t]

    m.charge_limit = Constraint(mp.time, rule=_ineq_charge_limit_pjm_pfp)


def ineq_discharge_limit_pjm_pfp(m):
    """Limits the energy discharged at each timestep to the device maximum."""
    mp = m.parent_block()

    def _ineq_discharge_limit_pjm_pfp(_m, t):
        return mp.Q_d_max >= mp.q_d[t] + mp.q_reg[t]

    m.discharge_limit = Constraint(mp.time, rule=_ineq_discharge_limit_pjm_pfp)


#############################
# MISO Pay-for-Performance ##
#############################


def eq_objective_miso_pfp(m):
    """Net revenue over the time horizon for pay-for-performance in MISO regulation market."""
    mp = m.parent_block()
    _expr = sum((mp.price_electricity[t] * mp.q_d[t] - mp.price_electricity[t] * mp.q_r[t]
                 + (1 + mp.Make_whole) * mp.q_reg[t] * mp.perf_score[t] * mp.price_regulation[t])
                * math.e ** (-t * mp.R) for t in mp.time)

    mp.objective_expr += _expr
    m.objective_rt = Expression(expr=_expr)


def eq_stateofcharge_miso_pfp(m):
    """Definition of state of charge for device in MISO market."""
    mp = m.parent_block()

    def _eq_stateofcharge_miso_pfp(_m, t):
        return mp.Self_discharge_efficiency * mp.s[t] + mp.Round_trip_efficiency * mp.q_r[t] \
                - mp.q_d[t] + mp.Round_trip_efficiency * mp.fraction_reg_down[t] * mp.q_reg[t] \
                - mp.fraction_reg_up[t] * mp.q_reg[t] == mp.s[t+1]

    m.stateofcharge = Constraint(mp.time, rule=_eq_stateofcharge_miso_pfp)


def ineq_charge_limit_miso_pfp(m):
    """Limits the energy charged at each timestep to the device maximum."""
    mp = m.parent_block()

    def _ineq_charge_limit_miso_pfp(_m, t):
        return mp.Q_r_max >= mp.q_r[t] + mp.q_reg[t]

    m.charge_limit = Constraint(mp.time, rule=_ineq_charge_limit_miso_pfp)


def ineq_discharge_limit_miso_pfp(m):
    """Limits the energy discharged at each timestep to the device maximum."""
    mp = m.parent_block()

    def _ineq_discharge_limit_miso_pfp(_m, t):
        return mp.Q_d_max >= mp.q_d[t] + mp.q_reg[t]

    m.discharge_limit = Constraint(mp.time, rule=_ineq_discharge_limit_miso_pfp)


#################################
#  ISO-NE Pay-for-Performance  ##
#            FW                ##
#################################

#//////////////////////////////////////////////////#
def eq_objective_isone_pfp(m):
    """Net revenue over the time horizon for pay-for-performance in ISO-NE market."""
    mp = m.parent_block()

    _expr = sum((mp.price_electricity[t] * mp.q_d[t] - mp.price_electricity[t] * mp.q_r[t]
                 + mp.q_reg[t] * (mp.mi_mult[t] * mp.price_reg_service[t] + mp.price_regulation[t]) * mp.perf_score[t]) * math.e ** (-t * mp.R) for t in mp.time)

    mp.objective_expr += _expr
    m.objective_rt = Expression(expr=_expr)
#//////////////////////////////////////////////////#

def eq_stateofcharge_isone_pfp(m):
    """Definition of state of charge for device in ISO-NE market."""
    mp = m.parent_block()

    def _eq_stateofcharge_isone_pfp(_m, t):
        return mp.Self_discharge_efficiency * mp.s[t] + mp.Round_trip_efficiency * mp.q_r[t] \
                - mp.q_d[t] + mp.Round_trip_efficiency * mp.fraction_reg_down[t] * mp.q_reg[t] \
                - mp.fraction_reg_up[t] * mp.q_reg[t] == mp.s[t+1]

    m.stateofcharge = Constraint(mp.time, rule=_eq_stateofcharge_isone_pfp)


#################################
#   NYISO Pay-for-Performance  ##
#            FW                ##
#################################
def eq_objective_nyiso_pfp(m):
    """Net revenue over the time horizon for pay-for-performance in the NYISO market."""
    mp = m.parent_block()

    # Expression without movement price because it makes no sense for the DAM????
    # _expr = sum((mp.price_electricity[t] * mp.q_d[t] - mp.price_electricity[t] * mp.q_r[t]
    #              + mp.price_electricity[t] * mp.fraction_reg_up[t] * mp.q_reg[t]
    #              - mp.price_electricity[t] * mp.fraction_reg_down[t] * mp.q_reg[t]
    #              + mp.q_reg[t] * mp.rmccp[t] * (1 - 1.1*(1-mp.Perf_score)) )
    #              * math.e ** (-t * mp.R) for t in mp.time)
    _expr = sum((mp.price_electricity[t] * mp.q_d[t] - mp.price_electricity[t] * mp.q_r[t]
                 + mp.price_electricity[t] * mp.fraction_reg_up[t] * mp.q_reg[t]
                 - mp.price_electricity[t] * mp.fraction_reg_down[t] * mp.q_reg[t]
                 + mp.q_reg[t] * mp.price_regulation[t] * (1 - 1.1*(1-mp.perf_score[t])) )
                 * math.e ** (-t * mp.R) for t in mp.time)

    # _expr = sum((mp.price_electricity[t] * mp.q_d[t] - mp.price_electricity[t] * mp.q_r[t]
    #              + mp.q_reg[t] * mp.rmccp[t] * (1 - 1.1*(1-mp.Perf_score)) )
    #              * math.e ** (-t * mp.R) for t in mp.time)

    # # Expression with movement price...
    # # TODO: need 'movement' from NYISO
    # _expr = sum((mp.price_electricity[t] * mp.q_d[t] - mp.price_electricity[t] * mp.q_r[t]
    #              + mp.q_reg[t] * mp.rmccp[t] * (1 - 1.1*(1-mp.Perf_score))
    #              + mp.rmpcp[t] * mp.movement[t] * mp.Perf_score) * math.e ** (-t * mp.R) for t in mp.time)

    mp.objective_expr += _expr
    m.objective_rt = Expression(expr=_expr)

def eq_stateofcharge_nyiso_pfp(m):
    """Definition of state of charge for device in NYISO market."""
    mp = m.parent_block()

    def _eq_stateofcharge_nyiso_pfp(_m, t):
        return mp.Self_discharge_efficiency * mp.s[t] + mp.Round_trip_efficiency * mp.q_r[t] \
               - mp.q_d[t] + mp.Round_trip_efficiency * mp.fraction_reg_down[t] * mp.q_reg[t] \
               - mp.fraction_reg_up[t] * mp.q_reg[t] == mp.s[t + 1]

    m.stateofcharge = Constraint(mp.time, rule=_eq_stateofcharge_nyiso_pfp)

#################################
#    SPP Pay-for-Performance   ##
#            FW                ##
#################################
def eq_objective_spp_pfp(m):
    """Net revenue over the time horizon for pay-for-performance in the SPP market."""
    mp = m.parent_block()

    # Expression without Mileage because there's no data for it in the SPP DA market
    # _expr = sum((mp.price_electricity[t] * mp.q_d[t] - mp.price_electricity[t] * mp.q_r[t]
    #              + mp.ru[t] * mp.q_ru[t] + mp.rd[t] * mp.q_rd[t]
    #              + mp.price_electricity[t] * mp.q_ru[t] * mp.fraction_reg_up[t]
    #              - mp.price_electricity[t] * mp.q_rd[t] * mp.fraction_reg_down[t])
    #             * math.e ** (-t * mp.R) for t in mp.time)
    _expr = sum((mp.price_electricity[t] * mp.q_d[t] - mp.price_electricity[t] * mp.q_r[t]
                 + mp.price_reg_up[t] * mp.q_ru[t] + mp.price_reg_down[t] * mp.q_rd[t]
                 + mp.price_electricity[t] * mp.q_ru[t] * mp.fraction_reg_up[t]
                 - mp.price_electricity[t] * mp.q_rd[t] * mp.fraction_reg_down[t])
                * math.e ** (-t * mp.R) for t in mp.time)

    mp.objective_expr += _expr
    m.objective_rt = Expression(expr=_expr)

def eq_stateofcharge_spp_pfp(m):
    """Definition of state of charge for device in SPP market."""
    mp = m.parent_block()
    # Equation (8) CAISO paper
    def _eq_stateofcharge_spp_pfp(_m, t):
        return mp.Self_discharge_efficiency * mp.s[t] + mp.Round_trip_efficiency * mp.q_r[t] \
               - mp.q_d[t] + mp.Round_trip_efficiency * mp.fraction_reg_down[t] * mp.q_rd[t] \
               - mp.fraction_reg_up[t] * mp.q_ru[t] == mp.s[t + 1]

    m.stateofcharge = Constraint(mp.time, rule=_eq_stateofcharge_spp_pfp)


#################################
#   CAISO Pay-for-Performance  ##
#            FW                ##
#################################
def eq_objective_caiso_pfp(m):
    """Net revenue over the time horizon for pay-for-performance in the CAISO market."""
    mp = m.parent_block()

    # Expression without Mileage because
    # TODO: to be updated when the new names are ready
    # _expr = sum((mp.price_electricity[t] * mp.q_d[t] - mp.price_electricity[t] * mp.q_r[t]
    #              + mp.ru[t] * mp.q_ru[t] + mp.rd[t] * mp.q_rd[t]
    #              + mp.price_electricity[t] * mp.q_ru[t] * mp.fraction_reg_up[t]
    #              - mp.price_electricity[t] * mp.q_rd[t] * mp.fraction_reg_down[t])
    #             * math.e ** (-t * mp.R) for t in mp.time)

    #  TODO: Uncomment below and delete above when new names are ready
    _expr = sum((mp.price_electricity[t] * mp.q_d[t] - mp.price_electricity[t] * mp.q_r[t]
                 + mp.price_reg_up[t] * mp.q_ru[t] + mp.price_reg_down[t] * mp.q_rd[t]
                 + mp.price_electricity[t] * mp.q_ru[t] * mp.fraction_reg_up[t]
                 - mp.price_electricity[t] * mp.q_rd[t] * mp.fraction_reg_down[t]
                 + mp.perf_score_ru[t] * mp.mi_mult_ru[t] * mp.price_reg_serv_up[t]
                 + mp.perf_score_rd[t] * mp.mi_mult_rd[t] * mp.price_reg_serv_down[t])
                * math.e ** (-t * mp.R) for t in mp.time)

    mp.objective_expr += _expr
    m.objective_rt = Expression(expr=_expr)

def eq_stateofcharge_caiso_pfp(m):
    """Definition of state of charge for device in CAISO market."""
    mp = m.parent_block()
    # Equation (8) CAISO paper
    def _eq_stateofcharge_caiso_pfp(_m, t):
        return mp.Self_discharge_efficiency * mp.s[t] + mp.Round_trip_efficiency * mp.q_r[t] \
               - mp.q_d[t] + mp.Round_trip_efficiency * mp.fraction_reg_down[t] * mp.q_rd[t] \
               - mp.fraction_reg_up[t] * mp.q_ru[t] == mp.s[t + 1]

    m.stateofcharge = Constraint(mp.time, rule=_eq_stateofcharge_caiso_pfp)


# def ineq_stateofcharge_minimum_isone_pfp(m): # WHY REPLICATE EQUATIONS THAT ARE THE SAME FOR DIFF MARKETS???? -FW
#     """
#     Requires the state of charge of the energy storage device to remain above the minimum at any given time.
#
#     :param m: Pyomo ConcreteModel object
#     :return: n/a
#     """
#     mp = m.parent_block()
#
#     def _ineq_stateofcharge_minimum_isone_pfp(_m, t):
#         try:
#             return mp.s[t] >= mp.S_min
#             # return mp.s[t] >= mp.S_min + mp.q_reg[t] # Other code supposedly so it can bid?
#         except ValueError:
#             return Constraint.Feasible
#
#     m.stateofcharge_minimum = Constraint(mp.time_interval, rule=_ineq_stateofcharge_minimum_isone_pfp)
#
#
# def ineq_stateofcharge_maximum_isone_pfp(m):
#     """
#     Requires the state of charge of the energy storage device to remain below the maximum at any given time.
#
#     :param m: Pyomo ConcreteModel object
#     :return: n/a
#     """
#     mp = m.parent_block()
#
#     def _ineq_stateofcharge_maximum_isone_pfp(_m, t):
#         try:
#             return mp.s[t] <= mp.S_max
#             # return mp.s[t] <= mp.S_max - mp.Gamma_c*mp.q_reg[t] # Other code supposedly so it can bid?
#         except ValueError:
#             return Constraint.Feasible
#
#     m.stateofcharge_maximum = Constraint(mp.time_interval, rule=_ineq_stateofcharge_maximum_isone_pfp)
#
#
# def ineq_charge_limit_isone_pfp(m):
#     """
#     Limits the energy charged at each timestep to the device maximum.
#
#     :param m: Pyomo ConcreteModel object
#     :return: n/a
#     """
#     mp = m.parent_block()
#
#     def _ineq_charge_limit_isone_pfp(_m, t):
#         return mp.Q_r_max >= mp.q_r[t] + mp.q_reg[t]
#
#     m.charge_limit = Constraint(mp.time_interval, rule=_ineq_charge_limit_isone_pfp)
#
#
# def ineq_discharge_limit_isone_pfp(m):
#     """
#     Limits the energy discharged at each timestep to the device maximum.
#
#     :param m: Pyomo ConcreteModel object
#     :return: n/a
#     """
#     mp = m.parent_block()
#
#     def _ineq_discharge_limit_isone_pfp(_m, t):
#         return mp.Q_d_max >= mp.q_d[t] + mp.q_reg[t]
#
#     m.discharge_limit = Constraint(mp.time_interval, rule=_ineq_discharge_limit_isone_pfp)

##################################
#          Generic              ##
#     Equality Constraints      ##
##################################

def eq_stateofcharge_initial(m):
    """Requires the initial state of charge of the energy storage device to equal its minimum value."""
    mp = m.parent_block()

    def _eq_stateofcharge_initial(_m, t):
        if not t == 0:
            return Constraint.Skip
        else:
            return mp.s[t] == mp.State_of_charge_init*mp.Energy_capacity

    m.stateofcharge_initial = Constraint(mp.soc_time, rule=_eq_stateofcharge_initial)

def eq_stateofcharge_final(m):
    """Requires the final state of charge of the energy storage device to equal its initial value."""
    mp = m.parent_block()

    def _eq_stateofcharge_final(_m, t):
        if not t == mp.soc_time[-1]:
            return Constraint.Skip
        else:
            return mp.s[t] == mp.State_of_charge_init*mp.Energy_capacity
        
    m.stateofcharge_final = Constraint(mp.soc_time, rule=_eq_stateofcharge_final)


##################################
#          Generic              ##
#   Inequality Constraints      ##
##################################


def ineq_stateofcharge_minimum(m):
    """Requires the state of charge of the energy storage device to remain above the minimum at any given time."""
    mp = m.parent_block()

    def _ineq_stateofcharge_minimum(_m, t):
        return mp.s[t] >= mp.State_of_charge_min*mp.Energy_capacity

    m.stateofcharge_minimum = Constraint(mp.soc_time, rule=_ineq_stateofcharge_minimum)


def ineq_stateofcharge_maximum(m):
    """Requires the state of charge of the energy storage device to remain below the maximum at any given time."""
    mp = m.parent_block()

    def _ineq_stateofcharge_maximum(_m, t):
        return mp.s[t] <= mp.State_of_charge_max*mp.Energy_capacity

    m.stateofcharge_maximum = Constraint(mp.soc_time, rule=_ineq_stateofcharge_maximum)


def ineq_stateofcharge_minimum_reserve_oneprod(m):
    """Requires the state of charge of the energy storage device to remain above the minimum at any given time. Accounts for penalty aversion parameters."""
    mp = m.parent_block()

    def _ineq_stateofcharge_minimum(_m, t):
        return mp.s[t+1] >= mp.Reserve_reg_min*mp.q_reg[t] + mp.State_of_charge_min*mp.Energy_capacity

    m.stateofcharge_minimum = Constraint(mp.time, rule=_ineq_stateofcharge_minimum)


def ineq_stateofcharge_maximum_reserve_oneprod(m):
    """Requires the state of charge of the energy storage device to remain below the maximum at any given time. Accounts for penalty aversion parameters."""
    mp = m.parent_block()

    def _ineq_stateofcharge_maximum(_m, t):
        return mp.s[t+1] <= mp.State_of_charge_max*mp.Energy_capacity- mp.Round_trip_efficiency*mp.Reserve_reg_max*mp.q_reg[t]

    m.stateofcharge_maximum = Constraint(mp.time, rule=_ineq_stateofcharge_maximum)


def ineq_stateofcharge_minimum_reserve_twoprod(m):
    """Requires the state of charge of the energy storage device to remain above the minimum at any given time. Accounts for penalty aversion parameters."""
    mp = m.parent_block()

    def _ineq_stateofcharge_minimum(_m, t):
        return mp.s[t+1] >= mp.Reserve_reg_min*mp.q_ru[t] + mp.State_of_charge_min*mp.Energy_capacity

    m.stateofcharge_minimum = Constraint(mp.time, rule=_ineq_stateofcharge_minimum)


def ineq_stateofcharge_maximum_reserve_twoprod(m):
    """Requires the state of charge of the energy storage device to remain below the maximum at any given time. Accounts for penalty aversion parameters."""
    mp = m.parent_block()

    def _ineq_stateofcharge_maximum(_m, t):
        return mp.s[t+1] <= mp.State_of_charge_max*mp.Energy_capacity - mp.Round_trip_efficiency*mp.Reserve_reg_max*mp.q_rd[t]

    m.stateofcharge_maximum = Constraint(mp.time, rule=_ineq_stateofcharge_maximum)


def ineq_power_limit_oneprod(m):
    """Limits the energy charged and discharged at each timestep to the device power rating."""
    mp = m.parent_block()

    def _ineq_power_limit_oneprod(_m, t):
        return mp.Power_rating >= mp.q_r[t] + mp.q_d[t] + mp.q_reg[t]

    m.power_limit = Constraint(mp.time, rule=_ineq_power_limit_oneprod)


def ineq_power_limit_twoprod(m):
    """Limits the energy charged and discharged at each timestep to the device power rating."""
    mp = m.parent_block()

    def _ineq_power_limit_twoprod(_m, t):
        return mp.Power_rating >= mp.q_r[t] + mp.q_d[t] + mp.q_ru[t] + mp.q_rd[t]

    m.power_limit = Constraint(mp.time, rule=_ineq_power_limit_twoprod)

##################################
#  Generic Pay-for-Performance  ##
#   Inequality Constraints      ##
##################################


def ineq_charge_limit_pfp(m):
    """Probably deprecated. Limits the energy charged at each timestep to the device maximum."""
    mp = m.parent_block()

    def _ineq_charge_limit_pfp(_m, t):
        return mp.Q_r_max >= mp.q_r[t] + mp.q_reg[t]

    m.charge_limit = Constraint(mp.time, rule=_ineq_charge_limit_pfp)


def ineq_discharge_limit_pfp(m):
    """Probably deprecated. Limits the energy discharged at each timestep to the device maximum."""
    mp = m.parent_block()

    def _ineq_discharge_limit_pfp(_m, t):
        return mp.Q_d_max >= mp.q_d[t] + mp.q_reg[t]

    m.discharge_limit = Constraint(mp.time, rule=_ineq_discharge_limit_pfp)


"""Market data processing tools."""
# from __future__ import absolute_import
# from xlrd.biffh import XLRDError

import pandas as pd
import numpy as np
import os
import calendar
import logging


def read_nyiso_data(fpath, year, month, nodeid, typedat="both", RT_DAM="both"):
    """
    Read NYISO historical LBMP, regulation capacity, and regulation movement prices and return NumPy ndarrays for those three prices.

    :param fpath: The path to the root of the NYISO data directory
    :type fpath: str
    :param year: Year of data to read
    :type year: int or str
    :param month: Month of data to read
    :type month: int or str
    :param nodeid: ID of the node to read
    :type nodeid: int or str
    :return: daLBMP, rtLBMP, daCAP, rtCAP, rtMOV: Hourly LBMP and regulation capacity/movement clearing price values.
    :rtype: NumPy ndarraysx
    """
    ############################################################

    daLBMP = np.empty([0])
    rtLBMP = np.empty([0])
    daCAP = np.empty([0])
    rtCAP = np.empty([0])
    rtMOV = np.empty([0])

    if isinstance(month, str):
        month = int(month)

    if isinstance(year, str):
        year = int(year)

    if isinstance(nodeid, str):
        nodeid = int(nodeid)

    ############################################################################################
    # folderfile = fpath
    # # TODO: path_nodes_file is a folder to adjust when integrating it to QuESt
    # path_nodes_file = 'C:/Users/fwilche/Documents/data_bank/NYISO/'
    path_nodes_file = "../_static/"
    pathf_nodeszones = os.path.join(fpath, path_nodes_file, "nodes_nyiso.csv")
    print(pathf_nodeszones)
    df_nodeszones = pd.read_csv(pathf_nodeszones, index_col=False)
    df_nodeszones_x = df_nodeszones.loc[df_nodeszones["Node ID"] == nodeid, :]

    if df_nodeszones_x.empty:
        print("The node does NOT exists in NYISO")
        # raise ValueError('Not a valid bus number!!!')
        return daLBMP, rtLBMP, daCAP, rtCAP, rtMOV
    else:
        if df_nodeszones_x.iloc[0, 0] == df_nodeszones_x.iloc[0, 2]:
            print("It's a zone node")
            zoneid = nodeid
            zone_gen = "zone"
        else:
            print("It's a gen node")
            zoneid = df_nodeszones_x.iloc[0, 2]
            zone_gen = "gen"
    print("Identified zone:")
    print(zoneid)
    ############################################################################################

    ndaysmonth = calendar.monthrange(year, month)
    ndaysmonth = int(ndaysmonth[1])

    for ix in range(ndaysmonth):
        day_x = ix + 1
        date_str = str(year) + str(month).zfill(2) + str(day_x).zfill(2)
        # print(date_str)

        fnameLBMP_DA = date_str + "damlbmp_" + zone_gen + ".csv"
        fnameASP_DA = date_str + "damasp.csv"

        fnameLBMP_RT = date_str + "realtime_" + zone_gen + ".csv"
        fnameASP_RT = date_str + "rtasp.csv"

        fname_path_LBMP_DA = os.path.join(
            fpath, "LBMP", "DAM", zone_gen, str(year), str(month).zfill(2), fnameLBMP_DA
        )
        fname_path_ASP_DA = os.path.join(
            fpath, "ASP", "DAM", str(year), str(month).zfill(2), fnameASP_DA
        )

        fname_path_LBMP_RT = os.path.join(
            fpath, "LBMP", "RT", zone_gen, str(year), str(month).zfill(2), fnameLBMP_RT
        )
        fname_path_ASP_RT = os.path.join(
            fpath, "ASP", "RT", str(year), str(month).zfill(2), fnameASP_RT
        )

        if typedat == "asp" or typedat == "both":
            # 20170201damasp.csv
            # 20180501rtasp.csv
            if RT_DAM == "RT" or RT_DAM == "both":
                try:
                    df_file = pd.read_csv(fname_path_ASP_RT, index_col=False)
                except FileNotFoundError:
                    rtCAP = np.empty([0])
                    rtMOV = np.empty([0])
                    logging.warning(
                        "read_nyiso_data: RT ASP file missing, returning empty array."
                    )
                    break

                if (
                    (year >= 2016 and month >= 6 and day_x >= 23)
                    or (year >= 2016 and month >= 7)
                    or (year >= 2017)
                ):
                    # NYCA Regulation Capacity ($/MWHr) - for newest type of data
                    df_file_rtCAP = df_file.loc[
                        df_file["PTID"] == zoneid, ["NYCA Regulation Capacity ($/MWHr)"]
                    ]
                    rtCAP = np.append(rtCAP, df_file_rtCAP.values)
                    # NYCA Regulation Movement ($/MW)
                    df_file_rtMOV = df_file.loc[
                        df_file["PTID"] == zoneid, ["NYCA Regulation Movement ($/MW)"]
                    ]
                    rtMOV = np.append(rtMOV, df_file_rtMOV.values)
                elif (
                    (year >= 2001 and month >= 10 and day_x >= 0)
                    or (year >= 2001 and month >= 11)
                    or (
                        year >= 2002
                        and not (
                            (year >= 2016 and month >= 6 and day_x >= 23)
                            or (year >= 2016 and month >= 7)
                            or (year >= 2017)
                        )
                    )
                ):
                    df_file_rtCAP = df_file["East Regulation ($/MWHr)"]
                    rtCAP = np.append(rtCAP, df_file_rtCAP.values)
                    df_file_rtMOV = df_file[" NYCA Regulation Movement ($/MW)"]
                    rtMOV = np.append(rtMOV, df_file_rtMOV.values)
                    # RT ancillary services for NYISO start on July 2004

            if RT_DAM == "DAM" or RT_DAM == "both":
                try:
                    df_file = pd.read_csv(fname_path_ASP_DA, index_col=False)
                except FileNotFoundError:
                    daCAP = np.empty([0])
                    logging.warning(
                        "read_nyiso_data: DA ASP file missing, returning empty array."
                    )
                    break

                if (
                    (year >= 2016 and month >= 6 and day_x >= 23)
                    or (year >= 2016 and month >= 7)
                    or (year >= 2017)
                ):
                    df_file_daCAP = df_file.loc[
                        df_file["PTID"] == zoneid, ["NYCA Regulation Capacity ($/MWHr)"]
                    ]
                    daCAP = np.append(daCAP, df_file_daCAP.values)
                elif (
                    (year >= 2001 and month >= 10 and day_x >= 0)
                    or (year >= 2001 and month >= 11)
                    or (
                        year >= 2002
                        and not (
                            (year >= 2016 and month >= 6 and day_x >= 23)
                            or (year >= 2016 and month >= 7)
                            or (year >= 2017)
                        )
                    )
                ):
                    df_file_daCAP = df_file["East Regulation ($/MWHr)"]
                    daCAP = np.append(daCAP, df_file_daCAP.values)
                else:
                    df_file_daCAP = df_file["Regulation ($/MWHr)"]
                    daCAP = np.append(daCAP, df_file_daCAP.values)

        if typedat == "lbmp" or typedat == "both":
            # 20170201damlbmp_gen.csv
            # 20170201damlbmp_zone.csv
            # 20170201realtime_gen.csv
            if RT_DAM == "RT" or RT_DAM == "both":
                try:
                    df_rtLBMP = pd.read_csv(fname_path_LBMP_RT, index_col=False)
                except FileNotFoundError:
                    rtLBMP = np.empty([0])
                    logging.warning(
                        "read_nyiso_data: RT LMP file missing, returning empty array."
                    )
                    break

                rtLBMP_node_x = df_rtLBMP.loc[
                    df_rtLBMP["PTID"] == nodeid, ["LBMP ($/MWHr)"]
                ]
                rtLBMP = np.append(rtLBMP, rtLBMP_node_x.values)
                if rtLBMP_node_x.empty:
                    return (
                        np.empty([0]),
                        np.empty([0]),
                        np.empty([0]),
                        np.empty([0]),
                        np.empty([0]),
                    )

            if RT_DAM == "DAM" or RT_DAM == "both":
                try:
                    df_daLBMP = pd.read_csv(fname_path_LBMP_DA, index_col=False)
                except FileNotFoundError:
                    daLBMP = np.empty([0])
                    logging.warning(
                        "read_nyiso_data: DA LMP file missing, returning empty array."
                    )
                    break

                daLBMP_node_x = df_daLBMP.loc[
                    df_daLBMP["PTID"] == nodeid, ["LBMP ($/MWHr)"]
                ]
                daLBMP = np.append(daLBMP, daLBMP_node_x.values)
                if daLBMP_node_x.empty:
                    return (
                        np.empty([0]),
                        np.empty([0]),
                        np.empty([0]),
                        np.empty([0]),
                        np.empty([0]),
                    )

    return daLBMP, rtLBMP, daCAP, rtCAP, rtMOV


"""Valuation optimizer class."""
from __future__ import division, print_function, absolute_import
from quest_eras.quest_library.utilities.pomo import optimizer
from quest_eras.quest_library.utilities.pomo.valuation.isone.constraints import (
    ExpressionsBlock,
)
import logging
import pyomo.environ as pom
import pandas as pd
import numpy as np


class ValuationOptimizerISONE(optimizer.Optimizer):
    """A framework wrapper class for creating Pyomo ConcreteModels for valuating energy storage in ISONE electricity market."""

    def __init__(
        self,
        price_electricity=None,  # Price of 1 MWh in energy market [$/MWh]
        price_regulation=None,  # Price of 1MW being available for regulation (both up and down) in regulation market [$/MW]
        price_reg_service=None,  # Price of 1MW being deployed for regulation (both up and down) in regulation market [$/MW]
        cost_charge=None,  # Cost for charging [$]
        cost_discharge=None,  # Cost for discharing [$]
        mileage_mult=None,  # Regulation mileage multiplication factor[-]
        perf_score=None,  # Performance score of following regulation signal[0,1]
        fraction_reg_up=None,  # Fraction of available capacity actually deployed for regulation up [0,1]
        fraction_reg_down=None,  # Fraction of available capacity actually deployed for regulation down [0,1]
        market_type="arbitrage",  # Market type, default = energy arbitrage
        solver="glpk",  # Optimization solver, default = glpk
    ):

        self._model = pom.ConcreteModel()
        self._market_type = market_type
        self._solver = solver

        self._expressions_block = None

        self._price_electricity = price_electricity

        self._price_regulation = price_regulation

        self._price_reg_service = price_reg_service

        self._cost_charge = cost_charge
        self._cost_discharge = cost_discharge

        self._mileage_mult = mileage_mult

        self._perf_score = perf_score

        self._fraction_reg_up = fraction_reg_up
        self._fraction_reg_down = fraction_reg_down

        self._results = None
        self._gross_revenue = None

    @property
    def price_electricity(self):
        """Price of 1 MWh in energy market [$/MWh]."""
        return self._price_electricity

    @price_electricity.setter
    def price_electricity(self, value):
        """Set value for price_electricity."""
        self._price_electricity = value

    @property
    def price_regulation(self):
        """Price of 1MW being available for regulation (both up and down) in regulation market [$/MW]."""
        return self._price_regulation

    @price_regulation.setter
    def price_regulation(self, value):
        """Set value for price_regulation."""
        self._price_regulation = value

    @property
    def price_reg_service(self):
        """Price of 1MW being deployed for regulation (both up and down) in regulation market [$/MW]."""
        return self._price_reg_service

    @price_reg_service.setter
    def price_reg_service(self, value):
        """Set value for price_reg_service."""
        self._price_reg_service = value

    @property
    def cost_charge(self):
        """Cost for charging [$]."""
        return self._cost_charge

    @cost_charge.setter
    def cost_charge(self, value):
        """Set value for cost_charge."""
        self._cost_charge = value

    @property
    def cost_discharge(self):
        """Cost for discharging [$]."""
        return self._cost_discharge

    @cost_discharge.setter
    def cost_discharge(self, value):
        """Set value for cost_discharge."""
        self._cost_discharge = value

    @property
    def mileage_mult(self):
        """Regulation mileage multiplication factor[-]."""
        return self._mileage_mult

    @mileage_mult.setter
    def mileage_mult(self, value):
        """Set value for mileage_mult."""
        self._mileage_mult = value

    @property
    def perf_score(self):
        """Performance-score of following regulation signal[0,1]."""
        return self._perf_score

    @perf_score.setter
    def perf_score(self, value):
        """Set value for perf_score."""
        self._perf_score = value

    @property
    def fraction_reg_up(self):
        """Fraction of available capacity actually deployed for regulation up [0,1]."""
        return self._fraction_reg_up

    @fraction_reg_up.setter
    def fraction_reg_up(self, value):
        """Set value for fraction_reg_up."""
        self._fraction_reg_up = value

    @property
    def fraction_reg_down(self):
        """Fraction of available capacity actually deployed for regulation down [0,1]."""
        return self._fraction_reg_down

    @fraction_reg_down.setter
    def fraction_reg_down(self, value):
        """Set value for fraction_reg_down."""
        self._fraction_reg_down = value

    @property
    def solver(self):
        """Optimization solver."""
        return self._solver

    @solver.setter
    def solver(self, value):
        """Setvalue for solver."""
        self._solver = value

    @property
    def expressions_block(self):
        """Contain model objectives and constraints."""
        return self._expressions_block

    @expressions_block.setter
    def expressions_block(self, value):
        """Set value for expressions_block."""
        self._expressions_block = value

    @property
    def market_type(self):
        """Market type, default = energy arbitrage."""
        return self._market_type

    @market_type.setter
    def market_type(self, value):
        """Set value to market_type."""
        if isinstance(value, str):
            self._market_type = value
        else:
            raise TypeError("market_type property must be of type str.")

    @property
    def results(self):
        """Pandas DataFrame containing results."""
        return self._results

    @results.setter
    def results(self, value):
        if isinstance(value, pd.DataFrame):
            self._results = value
        else:
            raise TypeError("results must be a Pandas DataFrame.")

    @property
    def gross_revenue(self):
        """Net revenue generated over the time period as solved for in the optimization."""
        return self._gross_revenue

    @gross_revenue.setter
    def gross_revenue(self, value):
        self._gross_revenue = value

    def _set_model_param(self):
        """Set the model params for the Pyomo ConcreteModel."""
        m = self.model

        # Check if params common to all formulations are set.
        if not hasattr(m, "Power_rating"):
            # Power rating; equivalently, the maximum energy charged in one hour [MW].
            logging.debug(
                "ValuationOptimizer: No Power_rating provided, setting default..."
            )
            m.Power_rating = 20

        if not hasattr(m, "R"):
            # Discount/interest rate [hour^(-1)].
            logging.debug("ValuationOptimizer: No R provided, setting default...")
            m.R = 0

        if not hasattr(m, "Energy_capacity"):
            # Energy capacity [MWh].
            logging.debug(
                "ValuationOptimizer: No Energy_capacity provided, setting default..."
            )
            m.Energy_capacity = 5

        if not hasattr(m, "Self_discharge_efficiency"):
            # Fraction of energy maintained over one time period.
            logging.debug(
                "ValuationOptimizer: No Self_discharge_efficiency provided, setting default..."
            )
            m.Self_discharge_efficiency = 1.00
        elif getattr(m, "Self_discharge_efficiency") > 1.0:
            logging.warning(
                "ValuationOptimizer: Self_discharge_efficiency provided is greater than 1.0, interpreting as percentage..."
            )
            m.Self_discharge_efficiency = m.Self_discharge_efficiency / 100

        if not hasattr(m, "Round_trip_efficiency"):
            # Fraction of input energy that gets stored over one time period.
            logging.debug(
                "ValuationOptimizer: No Round_trip_efficiency provided, setting default..."
            )
            m.Round_trip_efficiency = 0.85
        elif getattr(m, "Round_trip_efficiency") > 1.0:
            logging.warning(
                "ValuationOptimizer: Round_trip_efficiency provided is greater than 1.0, interpreting as percentage..."
            )
            m.Round_trip_efficiency = m.Round_trip_efficiency / 100

        if not hasattr(m, "Reserve_reg_min"):
            # Fraction of q_reg bid to increase state of charge minimum by.
            logging.debug(
                "ValuationOptimizer: No Reserve_reg_min provided, setting default..."
            )
            m.Reserve_reg_min = 0
        elif getattr(m, "Reserve_reg_min") > 1.0:
            logging.warning(
                "ValuationOptimizer: Reserve_reg_min provided is greater than 1.0, interpreting as percentage..."
            )
            m.Reserve_reg_min = m.Reserve_reg_min / 100

        if not hasattr(m, "Reserve_reg_max"):
            # Fraction of q_reg bid to decrease state of charge maximum by.
            logging.debug(
                "ValuationOptimizer: No Reserve_reg_max provided, setting default..."
            )
            m.Reserve_reg_max = 0
        elif getattr(m, "Reserve_reg_max") > 1.0:
            logging.warning(
                "ValuationOptimizer: Reserve_reg_max provided is greater than 1.0, interpreting as percentage..."
            )
            m.Reserve_reg_max = m.Reserve_reg_max / 100

        if not hasattr(m, "State_of_charge_min"):
            # Fraction of energy capacity representing the minimum state of charge.
            logging.debug(
                "ValuationOptimizer: No State_of_charge_min provided, setting default..."
            )
            m.State_of_charge_min = 0
        elif getattr(m, "State_of_charge_min") > 1.0:
            logging.warning(
                "ValuationOptimizer: State_of_charge_min provided is greater than 1.0, interpreting as percentage..."
            )
            m.State_of_charge_min = m.State_of_charge_min / 100

        if not hasattr(m, "State_of_charge_max"):
            # Fraction of energy capacity representing the maximum state of charge.
            logging.debug(
                "ValuationOptimizer: No State_of_charge_max provided, setting default..."
            )
            m.State_of_charge_max = 1
        elif getattr(m, "State_of_charge_max") > 1.0:
            logging.warning(
                "ValuationOptimizer: State_of_charge_max provided is greater than 1.0, interpreting as percentage..."
            )
            m.State_of_charge_max = m.State_of_charge_max / 100

        if not hasattr(m, "State_of_charge_init"):
            # Initial state of charge as a fraction of energy capacity.
            logging.debug(
                "ValuationOptimizer: No State_of_charge_init provided, setting default..."
            )
            m.State_of_charge_init = 0.5
        elif getattr(m, "State_of_charge_init") > 1.0:
            logging.warning(
                "ValuationOptimizer: State_of_charge_init provided is greater than 1.0, interpreting as percentage..."
            )
            m.State_of_charge_init = m.State_of_charge_init / 100

        try:
            if not getattr(m, "fraction_reg_up", None):
                logging.debug(
                    "ValuationOptimizer: No fraction_reg_up provided, setting default..."
                )
                m.fraction_reg_up = 0.25
        except ValueError:  # fraction_reg_up is array-like
            if np.isnan(m.fraction_reg_up).any():
                logging.debug(
                    "ValuationOptimizer: fraction_reg_up array-like provided has None values, setting default..."
                )
                m.fraction_reg_up = 0.25

        try:
            if not getattr(m, "fraction_reg_down", None):
                logging.debug(
                    "ValuationOptimizer: No fraction_reg_down provided, setting default..."
                )
                m.fraction_reg_down = 0.25
        except ValueError:  # fraction_reg_down is array-like
            if np.isnan(m.fraction_reg_down).any():
                logging.debug(
                    "ValuationOptimizer: fraction_reg_down array-like provided has None values, setting default..."
                )
                m.fraction_reg_down = 0.25

        # Converts fraction_reg_up and fraction_reg_down to arrays.
        try:
            m.fraction_reg_up[len(m.price_electricity) - 1]
        except TypeError:
            m.fraction_reg_up = np.array([m.fraction_reg_up] * len(m.price_electricity))

        try:
            m.fraction_reg_down[len(m.price_electricity) - 1]
        except TypeError:
            m.fraction_reg_down = np.array(
                [m.fraction_reg_down] * len(m.price_electricity)
            )

        try:
            if not getattr(m, "perf_score", None):
                logging.debug(
                    "ValuationOptimizer: No perf_score provided, setting default..."
                )
                m.perf_score = 0.95
        except ValueError:  # perf_score is array-like
            if np.isnan(m.perf_score).any():
                logging.debug(
                    "ValuationOptimizer: perf_score array-like provided has None values, setting default..."
                )
                m.perf_score = 0.95

        # Converts perf_score to array.
        try:
            # TODO: what if it is an array but longer than price of electricity??
            m.perf_score[len(m.price_electricity) - 1]
        except TypeError:
            m.perf_score = np.array([m.perf_score] * len(m.price_electricity))
        except IndexError:
            logging.warning(
                "ValuationOptimizer: A perf_score array was provided but is shorter than the price_electricity array."
            )
            raise (
                IncompatibleDataException(
                    "ValuationOptimizer: There was a mismatch in array sizes between perf_score and price_electricity."
                )
            )

    def _set_model_var(self):
        """Set the model vars for the Pyomo ConcreteModel."""
        m = self.model

        if not hasattr(m, "s"):

            def _s_init(_m, t):
                """Initialize energy storage device's state of charge [MWh]."""
                return m.State_of_charge_init * m.Energy_capacity

            m.s = pom.Var(m.soc_time, initialize=_s_init, within=pom.NonNegativeReals)

        if not hasattr(m, "q_r"):

            def _q_r_init(_m, t):
                """Initialize energy buy (chargey) from energy market [MWh]."""
                return 0.0

            m.q_r = pom.Var(m.time, initialize=_q_r_init, within=pom.NonNegativeReals)

        if not hasattr(m, "q_d"):

            def _q_d_init(_m, t):
                """Initialize energy sell (discharge) to energy market [MWh]."""
                return 0.0

            m.q_d = pom.Var(m.time, initialize=_q_d_init, within=pom.NonNegativeReals)

        if not hasattr(m, "q_reg"):

            def _q_reg_init(_m, t):
                """Initialize energy capacity available to regulation market [MWh(1 MW capacity available for 1 hour)]."""
                return 0.0

            m.q_reg = pom.Var(
                m.time, initialize=_q_reg_init, within=pom.NonNegativeReals
            )

    def instantiate_model(self):
        """Instantiate the Pyomo ConcreteModel and populates it with supplied time series data."""
        if not hasattr(self, "model"):
            self.model = pom.ConcreteModel()

        m = self.model

        try:
            m.time = pom.RangeSet(0, len(self.price_electricity) - 1)
            m.soc_time = pom.RangeSet(0, len(self.price_electricity))
        except TypeError:
            # self.price_electricity is of type 'NoneType'
            m.time = []
            m.soc_time = []

        m.price_electricity = self.price_electricity

        m.price_regulation = self.price_regulation

        m.price_reg_service = self.price_reg_service

        m.cost_charge = self.cost_charge
        m.cost_discharge = self.cost_discharge

        m.mi_mult = self.mileage_mult

        # If fraction_reg_up/fraction_reg_down are provided to the instance, set them in the ConcreteModel.
        if self.fraction_reg_up is not None:
            m.fraction_reg_up = self.fraction_reg_up
        if self.fraction_reg_down is not None:
            m.fraction_reg_down = self.fraction_reg_down

        if self.perf_score is not None:
            m.perf_score = self.perf_score

    def populate_model(self):
        """Populate the Pyomo ConcreteModel based on the specified market_type."""
        self.model.objective_expr = 0.0

        self._set_model_param()
        self._set_model_var()

        self.expressions_block = ExpressionsBlock(self.market_type)

        try:
            self.expressions_block.set_expressions(self.model)
        except IndexError:
            # Array-like object(s) do(es) not match the length of the price_electricity array-like.
            raise (
                IncompatibleDataException(
                    "At least one of the array-like parameter objects is not the expected length. (It should match the length of the price_electricity object.)"
                )
            )
        else:
            self.model.objective = pom.Objective(
                expr=self.model.objective_expr, sense=pom.maximize
            )

        # if self.model.objective.value == 0.0:
        #     # Detect constant objective function value.
        #     raise(IncompatibleDataException('The objective function was ill-formed, resulting in a constant objective function.'))

    def _process_results(self):
        """Process optimization results for further evaluation."""
        m = self.model

        t = m.time
        q_r = [m.q_r[n].value for n in m.time]
        q_d = [m.q_d[n].value for n in m.time]
        q_reg = [m.q_reg[n].value for n in m.time]
        soc = [m.s[n].value for n in m.time]
        price_electricity = [m.price_electricity[n] for n in m.time]

        run_results = {
            "time": t,
            "q_r": q_r,
            "q_d": q_d,
            "q_reg": q_reg,
            "state of charge": soc,
            "price of electricity": price_electricity,
        }

        if self.market_type == "pfp":
            rev_arb = np.cumsum(
                np.array(
                    [
                        m.price_electricity[t] * (m.q_d[t].value - m.q_r[t].value)
                        for t in m.time
                    ]
                )
            )
            rev_reg = np.cumsum(
                np.array(
                    [
                        (m.mi_mult[t] * m.price_reg_service[t] + m.price_regulation[t])
                        * m.perf_score[t]
                        * m.q_reg[t].value
                        for t in m.time
                    ]
                )
            )

            revenue = rev_arb + rev_reg

            run_results["rev_arb"] = rev_arb
            run_results["rev_reg"] = rev_reg
            run_results["revenue"] = revenue
        # //////////////////////////////////////////////////#
        else:
            rev_arb = np.cumsum(
                np.array(
                    [
                        m.price_electricity[t] * (m.q_d[t].value - m.q_r[t].value)
                        for t in m.time
                    ]
                )
            )
            rev_reg = np.cumsum(np.array([0 for t in m.time]))

            revenue = rev_arb + rev_reg

            run_results["rev_arb"] = rev_arb
            run_results["rev_reg"] = rev_reg
            run_results["revenue"] = revenue

        try:
            self.gross_revenue = revenue[-1]
        except IndexError:
            # Revenue is of length-0, likely due to no price_electricity array-like being given before solving.
            self.gross_revenue = 0

        self.results = pd.DataFrame(run_results)

    def get_results(self):
        """Return the decision variables and derived quantities in a DataFrame, plus the net revenue."""
        return self.results, self.gross_revenue


class BadParameterException(Exception):
    pass


class IncompatibleDataException(Exception):
    pass


if __name__ == "__main__":
    with open("valuation_optimizer.log", "w"):
        pass

    logging.basicConfig(
        filename="valuation_optimizer.log",
        format="[%(levelname)s] %(asctime)s: %(message)s",
        level=logging.INFO,
    )



from __future__ import absolute_import

from kivy.uix.settings import InterfaceWithSidebar, Settings, SettingItem
from kivy.uix.label import Label
from kivy.properties import ListProperty


class ESAppSettingsInterface(InterfaceWithSidebar):
    pass


class ESAppSettings(Settings):
    interface_cls = 'ESAppSettingsInterface'

    def __init__(self, **kwargs):
        super(ESAppSettings, self).__init__(**kwargs)
        #self.register_type('title', SettingsTitle)

    def add_kivy_panel(self):
        """
        Hides the Kivy settings.
        """
        pass


class SettingsItem(SettingItem):
    pass


class SettingsTitle(Label):
    title = Label.text


class SettingsBoolean(SettingsItem):
    values = ListProperty(['0', '1'])


from __future__ import absolute_import

import os
from functools import partial

from kivy.uix.screenmanager import Screen
from kivy.app import App

from es_gui.resources.widgets.common import WarningPopup, NavigationButton
from es_gui.proving_grounds.help_carousel import HelpCarouselModalView


class PerformanceHomeScreen(Screen):
    """
    The home screen for doing performance simulation energy storage analysis.
    """
    def __init__(self, **kwargs):
        super(PerformanceHomeScreen, self).__init__(**kwargs)

        # Initialize data management system.
#        self.dms = BtmDMS(
#            max_memory=App.get_running_app().config.getint('performance', 'performance_dms_size')*1000,
#            save_data=bool(App.get_running_app().config.getint('performance', 'performance_dms_save')),
#            save_name='performance_dms.p',
#            home_path='data',
#            )
#        self.handler.dms = self.dms
        
    def on_enter(self):
        ab = self.manager.nav_bar
        ab.reset_nav_bar()
        ab.set_title('Performance Applications')

        help_button = NavigationButton(
            text='help',
            on_release=self.open_help_carousel,
        )

        ab.action_view.add_widget(help_button)

        data_manager = App.get_running_app().data_manager

    def open_help_carousel(self, *args):
        """
        """
        help_carousel_view = HelpCarouselModalView()
        help_carousel_view.title.text = "QuESt Performance"

        slide_01_text = "QuESt Performance is an application with tools for analyzing the effects of heating and cooling on energy storage use cases."

        slide_02_text = "The performance simulation estimates the effects of parasitic loads on energy storage performance.\n\nYou will need the following to use this tool:\n* EnergyPlus software\n\n* Input file\n\n* Weather file\n\n* Charge/discharge profile\n\n* Battery parameters."

        slide_03_text = "EnergyPlus is a building simulation software used to model energy consumption developed by the Department of Energy Building Technologies Office. To download EnergyPlus, simply navigate to www.energyplus.net/downloads and select the appropriate package." 
        
        slide_04_text = "Once downloaded, move the software to the Quest directory and rename to 'energyplus'."
        
        slide_05_text = "An EnergyPlus input file describing a battery energy storage device in an uninsulated shipping container is provided. If you have another input file you would like to use, place it in the quest -> data -> idf directory."
        
        slide_06_text = "Weather files may be obtained in either of two ways. There is a large repository of weather files formatted for EnergyPlus found at www.energyplus.net/weather, which may be downloaded and placed under quest -> data -> weather -> location. NSRDB data may also be downloaded with the Quest Data Manager for a specific location and year."

        slide_07_text = "The battery charge/discharge profile may be used from a behind-the-meter or valuation run. If you would like to use your own data, place the file in the data directory under profile. Ensure the data file is formatted as shown."
        
        slide_08_text = "Battery parameters can be found or determined from nameplate values and the 1C discharge curve. See  T. A. Nguyen, D. A. Copp, R. H. Byrne, and B. R. Chalamala, “Market evaluation of energy storage systems incorporating technology-specific nonlinear models,” IEEE Transactions on Power Systems, vol. 34, no. 5, pp. 3706–3715, 2019. The default values correspond to an LG 18650 lithium-ion cell."

        slides = [
            (os.path.join("es_gui", "resources", "help_views", "performance", "performance_home.png"), slide_01_text),
            (os.path.join("es_gui", "resources", "help_views", "performance", "performance_data_select.png"), slide_02_text),
            (os.path.join("es_gui", "resources", "help_views", "performance", "eplus_download.png"), slide_03_text),
            (os.path.join("es_gui", "resources", "help_views", "performance", "Inkedquest_directory_LI.jpg"), slide_04_text),
            (os.path.join("es_gui", "resources", "help_views", "performance", "idf_dir.png"), slide_05_text),
            (os.path.join("es_gui", "resources", "help_views", "performance", "eplus_weather.png"), slide_06_text),
#            (os.path.join("es_gui", "resources", "help_views", "performance", "nsrdb_data.png"), slide_06_text),
            (os.path.join("es_gui", "resources", "help_views", "performance", "test_data.png"), slide_07_text),
            (os.path.join("es_gui", "resources", "help_views", "performance", "performance_param_select.png"), slide_08_text)
        ]

        help_carousel_view.add_slides(slides)
        help_carousel_view.open()
